"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
from jax import Array
import jax.numpy as jnp
import jax.random as jr
from jax.random import PRNGKey
from typing import Callable, List, Tuple

def evolve_trees(parent1: Array, 
                 parent2: Array, 
                 keys: Array, 
                 type: int, 
                 reproduction_probability: float, 
                 reproduction_functions: List[Callable]) -> Tuple[Array, Array]:
    """Applies reproduction function to pair of candidates.

    Parameters
    ----------
    parent1 : Array
        First parent candidate.
    parent2 : Array
        Second parent candidate.
    keys : Array
        Random keys.
    type : int
        Type of reproduction function to apply.
    reproduction_probability : float
        Probability of a tree to be mutated.
    reproduction_functions : List[Callable]
        Functions that can be applied for reproduction.

    Returns
    -------
    Tuple[Array, Array]
        Pair of reproduced candidates.
    """
    child0, child1 = jax.lax.switch(type, reproduction_functions, parent1, parent2, keys, reproduction_probability)

    return child0, child1

def tournament_selection(population: Array, 
                         ranks: Array,
                         key: PRNGKey, 
                         tournament_size: int, 
                         population_indices: Array) -> Array:
    """Selects a candidate for reproduction from a tournament.

    Parameters
    ----------
    population : Array
        Population of candidates.
    ranks : Array
        Ranks of candidates.
    key : PRNGKey
        Random key.
    tournament_size : int
        Size of the tournament.
    population_indices : Array
        Indices of the population.

    Returns
    -------
    Array
        Candidate that won the tournament.
    """
    tournament_key, winner_key = jr.split(key)
    indices = jr.choice(tournament_key, population_indices, shape=(tournament_size,), replace=False)

    tournament_ranks = ranks[indices] + 1

    winner_index = jr.choice(winner_key, indices, p=1/tournament_ranks)

    return population[winner_index]

def identify_non_dominated(metrics: Array) -> Array:
    """
    Identifies non-dominated solutions in a set using vectorized operations.
    
    Parameters
    ----------
    metrics : Array
        Array of shape (n_solutions, n_objectives) with metrics for each solution.
        
    Returns
    -------
    Array
        Boolean mask identifying non-dominated solutions.
    """

    # For each solution i, check if it's dominated by any other solution j
    
    # Reshape metrics for broadcasting. Shape: (n_solutions, 1, n_objectives)
    metrics_i = jnp.expand_dims(metrics, axis=1)
    
    # Shape: (1, n_solutions, n_objectives)
    metrics_j = jnp.expand_dims(metrics, axis=0)
    
    # Check if j dominates i: 
    # 1. j is better or equal in all objectives
    # 2. j is strictly better in at least one objective
    j_better_or_equal = (metrics_j <= metrics_i)
    j_strictly_better = (metrics_j < metrics_i)
    
    # For each pair (i,j), check if j dominates i
    j_dominates_i = jnp.all(j_better_or_equal, axis=2) & jnp.any(j_strictly_better, axis=2)
    
    # Don't compare solutions with themselves
    mask = ~jnp.eye(metrics.shape[0], dtype=bool)
    j_dominates_i = j_dominates_i & mask
    
    # A solution is non-dominated if it's not dominated by any other solution
    dominated_by_others = ~jnp.any(j_dominates_i, axis=1)
    
    return dominated_by_others

def nsga2(metrics: Array) -> Array:
    """
    Selects individuals for the next generation using NSGA-II without crowding distance.

    Parameters
    ----------
    metrics : Array
        Array of shape (population_size, n_objectives) with metrics for each individual.
    selection_size : int
        The number of individuals to select.

    Returns
    -------
    Array
        Indices of the selected individuals.
    """
    all_indices = jnp.ones(metrics.shape[0])

    def cond_fun(state):
        selected_count, remaining_indices, _ = state
        return jnp.sum(remaining_indices) > 0

    def body_fun(state):
        current_rank, remaining_indices, ranks = state
        remaining_metrics = jnp.where(remaining_indices[:, None], metrics, jnp.inf * jnp.ones_like(metrics))
        non_dominated = identify_non_dominated(remaining_metrics)
        
        ranks = jnp.where(non_dominated, current_rank * jnp.ones_like(ranks), ranks)
        
        remaining_indices = jnp.where(non_dominated, jnp.zeros_like(remaining_indices), remaining_indices)
        
        return current_rank + 1, remaining_indices, ranks

    initial_state = (0, all_indices, all_indices * 0)
    
    _, _, ranks = jax.lax.while_loop(cond_fun, body_fun, initial_state)

    return ranks

def evolve_population(population: Array, 
                      ranks: Array,
                      key: PRNGKey, 
                      reproduction_type_probabilities: Array, 
                      reproduction_probability: float, 
                      population_indices: Array, 
                      population_size: int, 
                      num_trees: int, 
                      tournament_size: int,
                      reproduction_functions: List[Callable]) -> Array:
    """Reproduces pairs of candidates to obtain a new population.

    Parameters
    ----------
    population : Array
        Population of candidates.
    ranks : Array
        Ranks of the candidates.
    key : PRNGKey
        Random key.
    reproduction_type_probabilities : Array
        Probability of each reproduction function to be applied.
    reproduction_probability : float
        Probability of a tree to be mutated.
    population_indices : Array
        Indices of the population.
    population_size : int
        Size of the population.
    num_trees : int
        Number of trees in a candidate.
    tournament_size : int
        Number of candidates that compete in a tournament.
    reproduction_functions : List[Callable]
        Functions that can be applied for reproduction.

    Returns
    -------
    Array
        Evolved population.
    """
    
    left_key, right_key, repro_key, evo_key = jr.split(key, 4)

    # Sample parents for reproduction
    left_parents = jax.vmap(tournament_selection, in_axes=[None, None, 0, None, None])(population, 
                                                                                             ranks,
                                                                                             jr.split(left_key, population_size//2), 
                                                                                             tournament_size,
                                                                                             population_indices)
    
    right_parents = jax.vmap(tournament_selection, in_axes=[None, None, 0, None, None])(population, 
                                                                                              ranks,
                                                                                              jr.split(right_key, population_size//2), 
                                                                                              tournament_size,
                                                                                              population_indices)
    # Sample which reproduction function to apply to the parents
    reproduction_type = jr.choice(repro_key, jnp.arange(3), shape=(population_size//2,), p=reproduction_type_probabilities)

    left_children, right_children = jax.vmap(evolve_trees, in_axes=[0, 0, 0, 0, None, None])(left_parents, 
                                                                                             right_parents, 
                                                                                             jr.split(evo_key, (population_size//2, num_trees, 2)), 
                                                                                             reproduction_type, 
                                                                                             reproduction_probability, 
                                                                                             reproduction_functions)
    
    evolved_population = jnp.where(ranks[:,None,None,None] == 0, population, jnp.concatenate([left_children, right_children], axis=0))

    return evolved_population

def migrate_population(receiver: Array, 
                       sender: Array, 
                       receiver_metrics: Array, 
                       sender_metrics: Array, 
                       receiver_ranks: Array,
                       sender_ranks: Array,
                       migration_size: int, 
                       population_indices: Array) -> Tuple[Array, Array]:
    """Unfit candidates from one population are replaced with fit candidates from another population.

    Parameters
    ----------
    receiver : Array
        Population that receives new candidates.
    sender : Array
        Population that sends fit candidates.
    receiver_fitness : Array
        Fitness of the candidates in the receiving population.
    sender_fitness : Array
        Fitness of the candidates in the sending population.
    receiver_ranks : Array
        Dominance ranks of candidates in the receiving population.
    sender_ranks : Array
        Dominance ranks of candidates in the sending population.
    migration_size : int
        How many candidates are migrated to new population.
    population_indices : Array
        Indices of the population.

    Returns
    -------
    Tuple[Array, Array, Array]
        Population after migration and the corresponding fitness and dominance ranks.
    """
    # Identify the Pareto front of the sender population
    sender_dominated_ranks = jnp.argsort(sender_ranks, descending=False)
    receiver_dominated_ranks = jnp.argsort(receiver_ranks, descending=True)

    # Replace the selected locations in the receiver with Pareto front individuals
    migrated_population = jnp.where((population_indices < migration_size)[:,None,None,None], sender[sender_dominated_ranks], receiver[receiver_dominated_ranks])
    migrated_metrics = jnp.where((population_indices < migration_size)[:,None], sender_metrics[sender_dominated_ranks], receiver_metrics[receiver_dominated_ranks])

    new_ranks = nsga2(migrated_metrics)

    return migrated_population, migrated_metrics, new_ranks

def evolve_populations(jit_evolve_population: Callable, 
                       populations: Array, 
                       metrics: Array, 
                       key: PRNGKey, 
                       current_generation: int, 
                       migration_period: int, 
                       migration_size: int, 
                       reproduction_type_probabilities: Array, 
                       reproduction_probabilities: Array) -> Array:
    """Evolves each population independently.

    Parameters
    ----------
    jit_evolve_population : Callable
        Function for evolving trees that is jittable and parallelizable.
    populations : Array
        Populations of candidates.
    fitness : Array
        Fitness of candidates.
    key : PRNGKey
        Random key.
    current_generation : int
        Current generation number.
    migration_period : int
        After how many generations migration occurs.
    migration_size : int
        How many candidates are migrated to new population.
    reproduction_type_probabilities : Array
        Probability of each reproduction function to be applied.
    reproduction_probabilities : Array
        Probability of a tree to be mutated.

    Returns
    -------
    Array
        Evolved populations.
    """
    num_populations, population_size, _, _, _ = populations.shape
    population_indices = jnp.arange(population_size)

    nsga_ranks = jax.vmap(nsga2)(metrics)

    # Migrate candidates between populations. The populations and fitnesses are rolled for circular migration.
    populations, metrics, nsga_ranks = jax.lax.cond((num_populations > 1) & (((current_generation+1)%migration_period) == 0), 
                                    jax.vmap(migrate_population, in_axes=[0, 0, 0, 0, 0, 0, None, None]), 
                                    jax.vmap(lambda receiver, sender, receiver_fitness, sender_fitness, receiver_ranks, sender_ranks, migration_size, population_indices: (receiver, receiver_fitness, receiver_ranks), 
                                             in_axes=[0, 0, 0, 0, 0, 0, None, None]), 
                                        populations, 
                                        jnp.roll(populations, 1, axis=0), 
                                        metrics, 
                                        jnp.roll(metrics, 1, axis=0), 
                                        nsga_ranks,
                                        jnp.roll(nsga_ranks, 1, axis=0), 
                                        migration_size, 
                                        population_indices)
    
    new_population = jit_evolve_population(populations, 
                                        nsga_ranks,
                                        jr.split(key, num_populations), 
                                        reproduction_type_probabilities, 
                                        reproduction_probabilities, 
                                        population_indices)
    return new_population


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
from jaxtyping import Array
import jax.numpy as jnp
from typing import Tuple, Callable
from kozax.fitness_functions.base_fitness_function import BaseFitnessFunction
import diffrax

class ODEFitnessFunction(BaseFitnessFunction):
    """
    Evaluator for candidates on symbolic regression tasks.

    Parameters
    ----------
    solver : :class:`diffrax.AbstractSolver`, optional
        Solver used for integration. Default is `diffrax.Euler()`.
    dt0 : float, optional
        Initial step size for integration. Default is 0.01.
    max_steps : int, optional
        The maximum number of steps that can be used in integration. Default is 16**4.
    stepsize_controller : :class:`diffrax.AbstractStepSizeController`, optional
        Controller for the stepsize during integration. Default is `diffrax.ConstantStepSize()`.

    Attributes
    ----------
    dt0 : float
        Initial step size for integration.
    MSE : Callable
        Function that computes the mean squared error.
    system : :class:`diffrax.ODETerm`
        ODE term of the drift function.
    solver : :class:`diffrax.AbstractSolver`
        Solver used for integration.
    stepsize_controller : :class:`diffrax.AbstractStepSizeController`
        Controller for the stepsize during integration.
    max_steps : int
        The maximum number of steps that can be used in integration.

    Methods
    -------
    __call__(candidate, data, tree_evaluator)
        Evaluates the candidate on a task.
    evaluate_time_series(candidate, x0, ts, ys, tree_evaluator)
        Integrate the candidate as a differential equation and compute the fitness given the predictions.
    drift(t, x, args)
        Drift function for the ODE system.
    """

    def __init__(self, solver: diffrax.AbstractSolver = diffrax.Euler(), dt0: float = 0.01, max_steps: int = 16**4, stepsize_controller: diffrax.AbstractStepSizeController = diffrax.ConstantStepSize()) -> None:
        self.dt0 = dt0
        self.MSE = lambda pred_ys, true_ys: jnp.mean(jnp.sum(jnp.abs(pred_ys - true_ys), axis=-1))/jnp.mean(jnp.abs(true_ys))
        self.system = diffrax.ODETerm(self.drift)
        self.solver = solver
        self.stepsize_controller = stepsize_controller
        self.max_steps = max_steps

    def __call__(self, candidate: Array, data: Tuple, tree_evaluator: Callable) -> float:
        """
        Evaluates the candidate on a task.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            The candidate solution to be evaluated.
        data : :class:`tuple`
            The data required to evaluate the candidate.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        float
            Fitness of the candidate.
        """
        x0, ts, ys = data
        fitness = jax.vmap(self.evaluate_time_series, in_axes=[None, 0, None, 0, None])(candidate, x0, ts, ys, tree_evaluator)
        return jnp.mean(fitness)
    
    def evaluate_time_series(self, candidate: Array, x0: Array, ts: Array, ys: Array, tree_evaluator: Callable) -> float:
        """
        Integrate the candidate as a differential equation and compute the fitness given the predictions.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            Candidate that is evaluated.
        x0 : :class:`jax.Array`
            Initial conditions of the environment.
        ts : :class:`jax.Array`
            Timepoints of which the system has to be solved.
        ys : :class:`jax.Array`
            Ground truth data used to compute the fitness.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        float
            Fitness of the candidate.
        """
        saveat = diffrax.SaveAt(ts=ts)
        sol = diffrax.diffeqsolve(
            self.system, self.solver, ts[0], ts[-1], self.dt0, x0, args=(candidate, tree_evaluator), saveat=saveat, max_steps=self.max_steps, stepsize_controller=self.stepsize_controller, 
            adjoint=diffrax.DirectAdjoint(), throw=False
        )
        pred_ys = sol.ys
        fitness = self.MSE(pred_ys, ys)
        return fitness
    
    def drift(self, t: float, x: Array, args: Tuple) -> Array:
        """
        Drift function for the ODE system.

        Parameters
        ----------
        t : float
            Current time.
        x : :class:`jax.Array`
            Current state.
        args : :class:`tuple`
            Additional arguments, including the candidate and tree evaluator.

        Returns
        -------
        :class:`jax.Array`
            Derivative of the state.
        """
        candidate, tree_evaluator = args
        dx = tree_evaluator(candidate, x)
        return dx


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
from kozax.environments.control_environments.control_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class HarmonicOscillator(EnvironmentBase):
    """
    Harmonic Oscillator environment for control tasks.

    Parameters
    ----------
    process_noise : float
        Standard deviation of the process noise.
    obs_noise : float
        Standard deviation of the observation noise.
    n_obs : int, optional
        Number of observations. Default is 2.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    n_var : int
        Number of variables in the state.
    n_control_inputs : int
        Number of control inputs.
    n_targets : int
        Number of targets.
    mu0 : :class:`jax.Array`
        Mean of the initial state distribution.
    P0 : :class:`jax.Array`
        Covariance matrix of the initial state distribution.
    q : float
        Process noise parameter.
    r : float
        Observation noise parameter.
    Q : :class:`jax.Array`
        Process noise covariance matrix.
    R : :class:`jax.Array`
        Observation noise covariance matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    sample_params(batch_size, mode, ts, key)
        Samples parameters for the environment.
    initialize_parameters(params, ts)
        Initializes the parameters of the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    fitness_function(state, control, target, ts)
        Computes the fitness function for the environment.
    cond_fn_nan(t, y, args, **kwargs)
        Checks for NaN or infinite values in the state.
    """

    def __init__(self, process_noise: float = 0.0, obs_noise: float = 0.0, n_obs: int = 2) -> None:
        self.n_dim = 1
        self.n_var = 2
        self.n_control_inputs = 1
        self.n_targets = 1
        self.mu0 = jnp.zeros(self.n_var)
        self.P0 = jnp.eye(self.n_var) * jnp.array([3.0, 1.0])
        super().__init__(process_noise, obs_noise, self.n_var, self.n_control_inputs, self.n_dim, n_obs)

        self.q = self.r = 0.5
        self.Q = jnp.array([[self.q, 0], [0, 0]])
        self.R = jnp.array([[self.r]])

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Tuple[Array, Array]:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        x0 : :class:`jax.Array`
            Initial states.
        targets : :class:`jax.Array`
            Target states.
        """
        init_key, target_key = jrandom.split(key)
        x0 = self.mu0 + jrandom.normal(init_key, shape=(batch_size, self.n_var)) @ self.P0
        targets = jrandom.uniform(target_key, shape=(batch_size, self.n_targets), minval=-3, maxval=3)
        return x0, targets
    
    def sample_params(self, batch_size: int, mode: str, ts: Array, key: jrandom.PRNGKey) -> Tuple[Array, Array]:
        """
        Samples parameters for the environment.

        Parameters
        ----------
        batch_size : int
            Number of parameters to sample.
        mode : str
            Mode for sampling parameters. Options are "Constant", "Different", "Changing".
        ts : :class:`jax.Array`
            Time steps.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        omegas : :class:`jax.Array`
            Sampled omega parameters.
        zetas : :class:`jax.Array`
            Sampled zeta parameters.
        """
        omega_key, zeta_key, args_key = jrandom.split(key, 3)
        if mode == "Constant":
            omegas = jnp.ones((batch_size))
            zetas = jnp.zeros((batch_size))
        elif mode == "Different":
            omegas = jrandom.uniform(omega_key, shape=(batch_size,), minval=0.0, maxval=2.0)
            zetas = jrandom.uniform(zeta_key, shape=(batch_size,), minval=0.0, maxval=1.5)
        elif mode == "Changing":
            decay_factors = jrandom.uniform(args_key, shape=(batch_size, 2), minval=0.98, maxval=1.02)
            init_omegas = jrandom.uniform(omega_key, shape=(batch_size,), minval=0.5, maxval=1.5)
            init_zetas = jrandom.uniform(zeta_key, shape=(batch_size,), minval=0.0, maxval=1.0)
            omegas = jax.vmap(lambda o, d, t: o * (d ** t), in_axes=[0, 0, None])(init_omegas, decay_factors[:, 0], ts)
            zetas = jax.vmap(lambda z, d, t: z * (d ** t), in_axes=[0, 0, None])(init_zetas, decay_factors[:, 1], ts)
        return omegas, zetas

    def initialize_parameters(self, params: Tuple[Array, Array], ts: Array) -> None:
        """
        Initializes the parameters of the environment.

        Parameters
        ----------
        params : tuple of :class:`jax.Array`
            Parameters to initialize.
        ts : :class:`jax.Array`
            Time steps.
        """
        omega, zeta = params
        self.A = jnp.array([[0, 1], [-omega, -zeta]])
        self.b = jnp.array([[0.0, 1.0]]).T
        self.G = jnp.array([[0, 0], [0, 1]])
        self.V = self.process_noise * self.G
        self.C = jnp.eye(self.n_var)[:self.n_obs]
        self.W = self.obs_noise * jnp.eye(self.n_obs)

    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        return self.A @ state + self.b @ args
    
    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V
    
    def fitness_function(self, state: Array, control: Array, target: Array, ts: Array) -> float:
        """
        Computes the fitness function for the environment.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        control : :class:`jax.Array`
            Control inputs.
        target : :class:`jax.Array`
            Target states.
        ts : :class:`jax.Array`
            Time steps.

        Returns
        -------
        float
            Fitness value.
        """
        x_d = jnp.array([jnp.squeeze(target), 0])
        u_d = -jnp.linalg.pinv(self.b) @ self.A @ x_d
        costs = jax.vmap(lambda _state, _u: (_state - x_d).T @ self.Q @ (_state - x_d) + (_u - u_d) @ self.R @ (_u - u_d))(state, control)
        return jnp.sum(costs)
    
    def cond_fn_nan(self, t: float, y: Array, args: Tuple, **kwargs) -> float:
        """
        Checks for NaN or infinite values in the state.

        Parameters
        ----------
        t : float
            Current time.
        y : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        float
            -1.0 if NaN or infinite values are found, 1.0 otherwise.
        """
        return jnp.where(jnp.any(jnp.isinf(y) + jnp.isnan(y)), -1.0, 1.0)


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

from jax.random import PRNGKey
import abc
from typing import Tuple, Any
from jaxtyping import Array

class EnvironmentBase(abc.ABC):
    """
    Abstract base class for time series environments in symbolic regression tasks.

    Parameters
    ----------
    n_var : int
        Number of variables in the state.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    terminate_event(state, **kwargs)
        Checks if the termination condition is met.
    """

    def __init__(self, n_var: int, process_noise: float) -> None:
        self.n_var = n_var
        self.process_noise = process_noise

    @abc.abstractmethod
    def sample_init_states(self, batch_size: int, key: PRNGKey) -> Any:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        Any
            Initial states.
        """
        raise NotImplementedError
    
    @abc.abstractmethod
    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        raise NotImplementedError

    def terminate_event(self, state: Array, **kwargs) -> bool:
        """
        Checks if the termination condition is met.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        kwargs : dict
            Additional arguments.

        Returns
        -------
        bool
            True if the termination condition is met, False otherwise.
        """
        return False


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
from jaxtyping import Array
import jax.numpy as jnp
import jax.random as jr
from typing import Tuple, Callable, Any
from kozax.fitness_functions.base_fitness_function import BaseFitnessFunction
import gymnax

class GymFitnessFunction(BaseFitnessFunction):
    """
    Evaluator for static symbolic policies in control tasks using Gym environments.

    Parameters
    ----------
    env_name : str
        Name of the Gym environment to be used.

    Attributes
    ----------
    env : gymnax environment
        The gymnax environment.
    env_params : dict
        Parameters for the gymnax environment.
    num_steps : int
        Number of steps in the environment for each episode.

    Methods
    -------
    __call__(candidate, keys, tree_evaluator)
        Evaluates the candidate on a task.
    evaluate_trajectory(candidate, key, tree_evaluator)
        Evaluates a rollout of the candidate in the environment.
    """

    def __init__(self, env_name: str) -> None:
        self.env, self.env_params = gymnax.make(env_name)
        self.num_steps = self.env_params.max_steps_in_episode

    def __call__(self, candidate: Array, keys: Array, tree_evaluator: Callable) -> float:
        """
        Evaluates the candidate on a task.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            The candidate solution to be evaluated.
        keys : :class:`jax.Array`
            Random keys for evaluation.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        float
            Fitness of the candidate.
        """
        reward = jax.vmap(self.evaluate_trajectory, in_axes=(None, 0, None))(candidate, keys, tree_evaluator)
        return jnp.mean(reward)
        
    def evaluate_trajectory(self, candidate: Array, key: jr.PRNGKey, tree_evaluator: Callable) -> Tuple[Array, float]:
        """
        Evaluates a rollout of the candidate in the environment.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            The candidate solution to be evaluated.
        key : :class:`jax.random.PRNGKey`
            Random key for evaluation.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        reward : float
            Total reward obtained during the trajectory.
        """
        key, subkey = jr.split(key)
        state, env_state = self.env.reset(subkey, self.env_params)

        def policy(state: Array) -> Array:
            """Symbolic policy."""
            a = tree_evaluator(candidate, state)
            return a

        def step_fn(carry: Tuple[Array, Any, jr.PRNGKey], _) -> Tuple[Tuple[Array, Any, jr.PRNGKey], Tuple[Array, float, bool]]:
            """Step function for lax.scan."""
            state, env_state, key = carry

            # Select action based on policy
            action = policy(state)

            # Step the environment
            key, subkey = jr.split(key)
            next_state, next_env_state, reward, done, _ = self.env.step(
                subkey, env_state, action, self.env_params
            )

            return (next_state, next_env_state, key), (state, reward, done)

        # Run the rollout using lax.scan
        (final_carry, (states, rewards, dones)) = jax.lax.scan(
            step_fn, (state, env_state, key), None, length=self.num_steps
        )
        
        return -jnp.sum(rewards)


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
import abc

_itemsize_kind_type = {
    (1, "i"): jnp.int8,
    (2, "i"): jnp.int16,
    (4, "i"): jnp.int32,
    (8, "i"): jnp.int64,
    (2, "f"): jnp.float16,
    (4, "f"): jnp.float32,
    (8, "f"): jnp.float64,
}

def force_bitcast_convert_type(val, new_type=jnp.int32):
    val = jnp.asarray(val)
    intermediate_type = _itemsize_kind_type[new_type.dtype.itemsize, val.dtype.kind]
    val = val.astype(intermediate_type)
    return jax.lax.bitcast_convert_type(val, new_type)

class EnvironmentBase(abc.ABC):
    def __init__(self, process_noise, obs_noise, n_var, n_control_inputs, n_dim, n_obs):
        self.process_noise = process_noise
        self.obs_noise = obs_noise
        self.n_var = n_var
        self.n_control_inputs = n_control_inputs
        self.n_dim = n_dim
        self.n_obs = n_obs

    @abc.abstractmethod
    def initialize_parameters(self, params, ts):
        raise NotImplementedError

    @abc.abstractmethod
    def sample_init_states(self, batch_size, key):
        raise NotImplementedError

    @abc.abstractmethod
    def sample_params(self, batch_size, mode, ts, key):
        raise NotImplementedError

    def f_obs(self, key, t_x):
        t, x = t_x
        new_key = jrandom.fold_in(key, force_bitcast_convert_type(t))
        out = self.C@x + jrandom.normal(new_key, shape=(self.n_obs,))@self.W
        return key, out
    
    @abc.abstractmethod
    def drift(self, t, state, args):
        raise NotImplementedError

    @abc.abstractmethod
    def diffusion(self, t, state, args):
        raise NotImplementedError

    @abc.abstractmethod
    def fitness_function(self, state, control, target, ts):
        raise NotImplementedError

    def terminate_event(self, state, **kwargs):
        return False


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
print("These device(s) are detected: ", jax.devices())
from jax import Array

import jax.numpy as jnp
import jax.random as jr
from jax.random import PRNGKey
import optax
from functools import partial
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, PartitionSpec as P, NamedSharding
from jax.experimental import mesh_utils
import sympy
from typing import Tuple, Callable
import time

import numpy as np

from kozax.genetic_operators.crossover import crossover_trees
from kozax.genetic_operators.initialization import sample_population, sample_tree
from kozax.genetic_operators.mutation import initialize_mutation_functions
from kozax.genetic_operators.reproduction import evolve_populations, evolve_population

# Function containers
def lambda_operator_arity1(f):
    return lambda x, y, _data: f(x)

def lambda_operator_arity2(f):
    return lambda x, y, _data: f(x, y)

def lambda_leaf(i):
    return lambda x, y, _data: _data[i]

class GeneticProgramming:
    """Genetic programming strategy of symbolic expressions.

    Parameters
    ----------
    num_generations : int
        The number of generations over which to evolve the population.
    population_size : int
        Number of candidates in the population.
    fitness_function : Callable
        Function that evaluates a candidate and assigns a fitness.
    operator_list : list, optional
        List of operators that can be included in the trees.
    variable_list : list, optional
        List of variables that can be included in the trees, which can vary between layers of a candidate.
    layer_sizes : Array, optional
        Size of each layer in a candidate.
    num_populations : int, optional
        Number of subpopulations.
    max_init_depth : int, optional
        Highest depth of a tree during initialization.
    max_nodes : int, optional
        Maximum number of nodes in a tree.
    device_type : str, optional
        Type of device on which the evaluation and evolution takes place.
    constant_sd : float, optional
        Standard deviation to sample constants.
    migration_period : int, optional
        Number of generations after which populations are migrated.
    migration_size : float, optional
        Number of candidates to migrate.
    tournament_size: int, optional
        Size of the tournament.
    constant_optimization_method : str, optional
        Method for optimizing constants. Options are "evolution", "gradient", or None.
    constant_optimization_N_offspring : int, optional
        Number of offspring for Evolution Strategies.
    constant_optimization_steps : int, optional
        The number of iterations to optimize the constants.
    start_constant_optimization : int, optional
        After which generation the optimization of constants should start.
    optimize_constants_elite : int, optional
        The number of the best candidates of which the constants are optimized.
    optimizer_class : optional
        optimizer for constant optimization.
    constant_step_size_init : float, optional
        Initial step size for the optimizer.
    constant_step_size_decay : float, optional
        Decay rate for the step size.
    max_fitness : float, optional
        Maximum fitness value.
    reproduction_probability_factors : Tuple[float], optional
        The reproduction probability for each subpopulation.
    crossover_probability_factors : Tuple[float], optional
        The crossover probability for each subpopulation.
    mutation_probability_factors : Tuple[float], optional
        The mutation probability for each subpopulation.
    sample_probability_factors : Tuple[float], optional
        The probability to sample a new candidate for each subpopulation.
    """

    def __init__(self, 
                 num_generations: int, 
                 population_size: int, 
                 fitness_function: Callable, 
                 operator_list: list = None,
                 variable_list: list = None,
                 layer_sizes: Array = jnp.array([1]),
                 num_populations: int = 1,
                 max_init_depth: int = 4,
                 max_nodes: int = 15,
                 device_type: str = 'cpu',
                 complexity_objective: bool = False,
                 constant_sd: float = 1.0,
                 migration_period: int = 5,
                 migration_size: float = 10,
                 tournament_size: int = 7,
                 constant_optimization_method: str = None,
                 constant_optimization_N_offspring: int = 50,
                 constant_optimization_steps: int = 1,
                 start_constant_optimization: int = 0,
                 optimize_constants_elite: int = 100,
                 optimizer_class = optax.adam,
                 constant_step_size_init: float = 0.1,
                 constant_step_size_decay: float = 0.99,
                 max_fitness: float = 1e8,
                 reproduction_probability_factors: float | Tuple[float] = (0.75, 0.75),
                 crossover_probability_factors: float | Tuple[float] = (0.9, 0.1),
                 mutation_probability_factors: float | Tuple[float] = (0.1, 0.9),
                 sample_probability_factors: float | Tuple[float] = (0.0, 0.0)) -> None:
        
        self.layer_sizes = layer_sizes
        assert num_populations > 0, "The number of populations should be larger than 0"
        self.num_populations = num_populations
        assert population_size > 0 and population_size % 2 == 0, "The population_size should be larger than 0 and an even number"
        self.population_size = population_size
        assert max_init_depth > 0, "The max initial depth should be larger than 0"
        self.max_init_depth = max_init_depth
        assert max_nodes > 0, "The max number of nodes should be larger than 0"
        self.max_nodes = max_nodes
        self.num_trees = jnp.sum(self.layer_sizes).item()
        assert self.num_trees > 0, "The number of trees should be larger than 0"
        assert tournament_size > 1, "The tournament size should be larger than 1"
        self.tournament_size = tournament_size

        assert num_generations > 0, "The number of generations should be larger than 0"
        self.num_generations = num_generations

        self.complexity_objective = complexity_objective
        assert constant_sd > 0, "The standard deviation of the constants should be larger than 0"
        self.constant_sd = constant_sd

        assert migration_period > 1, "The migration period should be larger than 1"
        self.migration_period = migration_period
        assert isinstance(migration_size, int), "The migration size should be an integer"
        self.migration_size = migration_size
        
        if isinstance(crossover_probability_factors, float):
            crossover_probability_factors = (crossover_probability_factors, crossover_probability_factors)
        assert (crossover_probability_factors[0] >= 0) & (crossover_probability_factors[1] >= 0), "The crossover probability should not be negative"

        if isinstance(mutation_probability_factors, float):
            mutation_probability_factors = (mutation_probability_factors, mutation_probability_factors)
        assert (mutation_probability_factors[0] >= 0) & (mutation_probability_factors[1] >= 0), "The mutation probability should not be negative"

        if isinstance(sample_probability_factors, float):
            sample_probability_factors = (sample_probability_factors, sample_probability_factors)
        assert (sample_probability_factors[0] >= 0) & (sample_probability_factors[1] >= 0), "The sample probability should not be negative"

        self.reproduction_type_probabilities = jnp.vstack([jnp.linspace(*crossover_probability_factors, self.num_populations),
                                                           jnp.linspace(*mutation_probability_factors, self.num_populations),
                                                           jnp.linspace(*sample_probability_factors, self.num_populations)]).T
        
        if isinstance(reproduction_probability_factors, float):
            reproduction_probability_factors = (reproduction_probability_factors, reproduction_probability_factors)
        assert (reproduction_probability_factors[0] > 0) & (reproduction_probability_factors[1] > 0), "The reproduction probability should be larger than 0"

        self.reproduction_probabilities = jnp.linspace(*reproduction_probability_factors, self.num_populations)

        self.max_fitness = max_fitness
        self.complexities = jnp.arange(self.num_trees * self.max_nodes)

        # Create a mapping from indices in breadth first space to depth first space
        self.map_b_to_d = self.map_breadth_indices_to_depth_indices(jnp.maximum(self.max_init_depth, 3))

        # Define general tree structures
        self.tree_indices = jnp.tile(jnp.arange(self.max_nodes)[:, None], reps=(1, 4))
        self.empty_tree = jnp.tile(jnp.array([0.0, -1.0, -1.0, 0.0]), (self.max_nodes, 1))

        self.initialize_node_library(operator_list, variable_list)

        print(f"Input data should be formatted as: {[self.node_to_string[i.item()] for i in self.variable_indices]}.")

        # Define jittable reproduction functions
        self.sample_args = (self.variable_indices, 
                            self.operator_indices, 
                            self.operator_probabilities, 
                            self.slots, 
                            self.constant_sd, 
                            self.map_b_to_d)
                
        self.sample_tree = partial(sample_tree,
                                   max_nodes=self.max_nodes, 
                                   tree_size=2**jnp.maximum(self.max_init_depth, 3) - 1,
                                   args=self.sample_args)
        
        self.sample_population = partial(sample_population, 
                                         num_trees=self.num_trees, 
                                         max_init_depth=self.max_init_depth, 
                                         variable_array=self.variable_array,
                                         sample_function=self.sample_tree)

        self.mutate_args = (self.sample_tree, 
                            self.max_nodes, 
                            self.max_init_depth, 
                            self.variable_indices, 
                            self.operator_indices, 
                            self.operator_probabilities, 
                            self.slots, 
                            self.constant_sd)
        
        self.mutate_trees = initialize_mutation_functions(self.mutate_args)

        self.partial_crossover = partial(crossover_trees, 
                                         operator_indices=self.operator_indices, 
                                         max_nodes=self.max_nodes)

        self.reproduction_functions = [self.partial_crossover, self.mutate_pair, self.sample_pair]

        self.jit_evolve_population = jax.jit(jax.vmap(partial(evolve_population, 
                                                     reproduction_functions=self.reproduction_functions, 
                                                     num_trees=self.num_trees, 
                                                     population_size=population_size,
                                                     tournament_size=self.tournament_size),
                                                    in_axes=[0, 0, 0, 0, 0, None]))
        
        self.jit_simplify_constants = jax.jit(jax.vmap(jax.vmap(jax.vmap(self.simplify_constants))))

        # Define partial fitness function for evaluation
        self.jit_evaluate_row_from_tree = partial(self.evaluate_row_from_tree, node_function_list=self.node_function_list)
        self.partial_fitness_function = lambda constants, nodes, data: fitness_function(jnp.concatenate([nodes, constants], axis=-1), data, self.tree_evaluator)

        # Define parallel evaluation functions
        self.vmap_trees = jax.vmap(self.partial_fitness_function, in_axes=[0, 0, None])
        self.vmap_gradients = jax.vmap(jax.value_and_grad(self.partial_fitness_function), in_axes=[0, 0, None])

        assert device_type in ["cpu", "gpu", "tpu"], "The device type is not supported"
        devices = mesh_utils.create_device_mesh((len(jax.devices(device_type)),))
        self.mesh = Mesh(devices, axis_names=('i'))
        self.data_mesh = NamedSharding(self.mesh, P())
        
        # Define hyperparameters for constant optimization
        self.constant_step_size_init = constant_step_size_init
        self.constant_step_size_decay = constant_step_size_decay
        self.optimize_constants_elite = optimize_constants_elite
        self.start_constant_optimization = start_constant_optimization
        self.optimizer_class = optimizer_class

        assert constant_optimization_method in ["evolution", "gradient", None], "This optimization method is not implemented"
        if constant_optimization_method == "evolution":
            assert constant_optimization_N_offspring > 0, "The offspring size for constant optimization should be larger than 0"
            self.constant_optimization_steps = constant_optimization_steps
            self.optimize_constants_function = jax.vmap(partial(self.optimize_constants_with_evolution, n_iterations=constant_optimization_steps), in_axes=[0, None, 0, None])
            self.n_offspring = constant_optimization_N_offspring
            self.constant_optimization = True
        elif constant_optimization_method == "gradient":
            self.optimize_constants_function = partial(self.optimize_constants_with_gradients, n_epoch=constant_optimization_steps)
            self.constant_optimization = True
            
        else:
            self.optimize_constants_function = None
            self.constant_optimization = False

        # Define sharded functions for evaluation and optimization
        @partial(shard_map, mesh=self.mesh, in_specs=(P('i'), P(None)), out_specs=P('i'), check_rep=False)
        def shard_eval(array, data):
            fitness = self.vmap_trees(array[..., 3:], array[..., :3], data)

            # Regularize invalid solutions
            nan_or_inf = jax.vmap(lambda f: jnp.isinf(f) + jnp.isnan(f))(fitness)
            fitness = jnp.where(nan_or_inf, jnp.ones(fitness.shape) * self.max_fitness, fitness)
            
            return jnp.minimum(fitness, self.max_fitness * jnp.ones_like(fitness))
            
        @partial(shard_map, mesh=self.mesh, in_specs=(P('i'), P(None), P('i'), P()), out_specs=(P('i'), P('i')), check_rep=False)
        def shard_optimize(array, data, keys, step_size):
            result, _array = self.optimize_constants_function(array, data, keys, step_size)
            return result, _array
        
        self.jit_eval = jax.jit(shard_eval)
        self.jit_optimize = jax.jit(shard_optimize)

        self.reset()
        
    def _warm_up_jit_functions(self, dummy_population: Array, data: Tuple) -> None:
        """
        Warm up JIT compilation with actual data shapes.

        Parameters
        ----------
        dummy_population : Array
            A JAX array representing a sample population with the same
            shape as will be used in actual evolution.
        data : Tuple
            A tuple containing the data used for evaluating fitness functions. This method prints the compilation start and completion time as a diagnostic.

        Returns
        -------
        None            
        """

        print("Compiling code for evaluation and evolution...")
        start = time.time()

        dummy_key = jr.PRNGKey(0)
        
        # Use actual data for compilation
        flat_populations = dummy_population.reshape(self.num_populations * self.population_size, *dummy_population.shape[2:])
        dummy_fitness, dummy_population = self.evaluate_population(dummy_population, data, dummy_key)

        self.jit_simplify_constants(dummy_population)
        
        # Warm up evolve function
        dummy_population = self.evolve_population(dummy_population, dummy_fitness, dummy_key)
        jax.block_until_ready(dummy_population)
        print(f"Finished compilation in {(time.time() - start):.2f} seconds")

    def map_breadth_indices_to_depth_indices(self, depth: int) -> Array:
        """
        Creates a mapping from the breadth-first index to depth-first index given a depth.

        Parameters
        ----------
        depth : int
            Depth of the tree.

        Returns
        -------
        Array
            Index mapping.
        """
        max_nodes = 2**depth - 1
        current_depth = 0
        map_b_to_d = jnp.zeros(max_nodes)

        for i in range(max_nodes):
            if i > 0:
                parent = (i + (i % 2) - 2) // 2  # Determine parent position
                value = map_b_to_d[parent]
                if (i % 2) == 0:  # Right child
                    new_value = value + 2**(depth - current_depth + 1)
                else:  # Left child
                    new_value = value + 1
                map_b_to_d = map_b_to_d.at[i].set(new_value)
            current_depth += i == (2**current_depth - 1)  # If last node at current depth is reached, increase current depth

        return max_nodes - 1 - map_b_to_d  # Inverse the mapping

    def initialize_node_library(self, operator_list: list, variable_list: list) -> None:
        """
        Initializes the node library.

        Parameters
        ----------
        operator_list : list
            List of operators that can be included in the trees.
        variable_list : list
            List of variables that can be included in the trees, which can vary between layers of a candidate.
        """
        string_to_node = {}  # Maps string to node
        node_to_string = {}
        node_function_list = [lambda x, y, _data: 0.0, lambda x, y, _data: 0.0]

        if operator_list is None:
            operator_list = [("+", lambda x, y: jnp.add(x, y), 2, 0.1), 
                             ("-", lambda x, y: jnp.subtract(x, y), 2, 0.1),
                             ("*", lambda x, y: jnp.multiply(x, y), 2, 0.1),
                             ("/", lambda x, y: jnp.divide(x, y), 2, 0.1),
                             ("**", lambda x, y: jnp.power(x, y), 2, 0.1)
                             ]
            
        if variable_list is None:
            assert len(self.layer_sizes) == 1, "If more than one type of tree are optimized, you have to specify the input variables"
            variable_list = [["x" + str(i) for i in range(self.layer_sizes[0])]]

        n_operands = [0, 0] #Add 0 for empty node and constant node
        index = 2
        operator_probabilities = jnp.zeros(len(operator_list))

        assert len(operator_list) > 0, "No operators were given"

        for operator_tuple in operator_list:
            string = operator_tuple[0]
            f = operator_tuple[1] #Executable function
            arity = operator_tuple[2] #Number of children
            if len(operator_tuple) == 4:
                probability = operator_tuple[3] #Probability of function to be sampled
            else:
                probability = 0.1

            if string not in string_to_node:
                string_to_node[string] = index
                node_to_string[index] = string
                if arity == 1:
                    node_function_list.append(lambda_operator_arity1(f))
                    n_operands.append(1)
                elif arity == 2:
                    node_function_list.append(lambda_operator_arity2(f))
                    n_operands.append(2)
                operator_probabilities = operator_probabilities.at[index - 2].set(probability)
                index += 1

        self.operator_probabilities = operator_probabilities
        self.operator_indices = jnp.arange(2, index) #Store the indices corresponding to operator nodes
        var_start_index = index #The leaf nodes are appended to the operator list

        data_index = 0
        assert len(self.layer_sizes) == len(variable_list), "There is not a set of expressions for every type of layer"

        for var_list in variable_list:
            assert len(var_list) > 0, "An empty set of variables was given"
            for var_or_tuple in var_list:
                if isinstance(var_or_tuple, str): #Variables may be provided with or without probability
                    var = var_or_tuple
                else:
                    var = var_or_tuple[0]
                if var not in string_to_node:
                    try:
                        sympy.parsing.sympy_parser.parse_expr(f"1.0*{var}"), "Variable is not a valid expression"
                        string_to_node[var] = index
                        node_to_string[index] = var
                        node_function_list.append(lambda_leaf(data_index))
                        n_operands.append(0) #Leaf nodes have no children
                        index += 1
                        data_index += 1
                    except:
                        raise ValueError(f"Variable {var} is not a valid expression. Please use a valid expression or remove it from the variable list.")
        
        self.variable_indices = jnp.arange(var_start_index, index) #Store the indices corresponding to leaf nodes
        variable_array = jnp.zeros((self.num_trees, data_index))

        counter = 0
        for layer_i, var_list in enumerate(variable_list):
            p = jnp.zeros((data_index))
            for var_or_tuple in var_list:
                if isinstance(var_or_tuple, str): #Variables may be provided with or without probability
                    var = var_or_tuple
                    var_p = 0.1
                else:
                    var = var_or_tuple[0]
                    var_p = var_or_tuple[1]
                p = p.at[string_to_node[var] - var_start_index].set(var_p)

            for _ in range(self.layer_sizes[layer_i]):
                variable_array = variable_array.at[counter].set(p)
                counter += 1

        self.slots = jnp.array(n_operands)
        self.string_to_node = string_to_node
        self.node_to_string = node_to_string
        self.node_function_list = node_function_list
        self.variable_array = variable_array

    def fit(self, key: PRNGKey, data: Tuple, verbose: int = 0, save_pareto_front: bool = False, path_to_file: str = None) -> None:
        """
        Fits the genetic programming algorithm to the data.

        Parameters
        ----------
        data : Tuple
            Data required for evaluation.
        key : PRNGKey
            Random key.
        verbose : int, optional
            Whether to print the best fitness and solution at each generation. By choosing a positive integer, you can control the frequency of the output.
        save_pareto_front : bool, optional
            Whether to save the Pareto front to a file.
        path_to_file : str, optional
            Name of the file to save the Pareto front.
        """

        # Check if path exists when saving is requested
        if save_pareto_front:
            assert path_to_file is not None, "A file name must be provided when saving the Pareto front"
            import os
            # Check if directory exists and create it if needed
            directory = os.path.dirname(path_to_file)
            if directory and not os.path.exists(directory):
                os.makedirs(directory)

        key, init_key = jr.split(key)

        self.reset()
        
        # Warm up JIT functions with actual data shapes
        dummy_population = self.initialize_population(init_key)
        self._warm_up_jit_functions(dummy_population, data)

        population = dummy_population

        for g in range(self.num_generations):
            key, eval_key, sample_key = jr.split(key, 3)
            fitness, population = self.evaluate_population(population, data, eval_key)

            if verbose:
                if g % verbose == 0:
                    print(f"In generation {g+1}")
                    self.print_pareto_front()

            if g < (self.num_generations-1):
                population = self.evolve_population(population, fitness, sample_key)

        print("Final pareto front:")
        self.print_pareto_front(save_pareto_front, path_to_file)
        # print(eqx.debug.get_num_traces(self.jit_eval))

    def reset(self) -> None:
        """Resets the state of the genetic programming algorithm."""
        self.current_generation = 0
        self.constant_step_size = self.constant_step_size_init

        # The Pareto front keeps track of the best solutions at each complexity level
        self.pareto_front = (jnp.array([self.max_fitness]), jnp.ones((1, self.num_trees, self.max_nodes, 4)))        

    def initialize_population(self, key: PRNGKey) -> Array:
        """
        Randomly initializes the population.

        Parameters
        ----------
        key : PRNGKey
            Random key.

        Returns
        -------
        Array
            Population.
        """
        keys = jr.split(key, self.num_populations)
        populations = jax.vmap(self.sample_population, in_axes=[0, None])(keys, self.population_size)

        return self.jit_simplify_constants(populations)

    def evolve_population(self, populations: Array, fitness: Array, key: PRNGKey) -> Array:
        """
        Evolves each population independently.

        Parameters
        ----------
        populations : Array
            Populations of candidates.
        fitness : Array
            Fitness of candidates.
        key : PRNGKey
            Random key.

        Returns
        -------
        Array
            Evolved populations.
        """

        populations, fitness = jax.vmap(self.punish_duplicates)(populations, fitness) #Give duplicate candidates poor fitness

        fitness = jnp.expand_dims(fitness, axis=2) #Expand fitness to match the shape of the populations
        
        if self.complexity_objective:
            complexities = jax.vmap(lambda population: jax.vmap(lambda candidate: jnp.sum(candidate[:,:,0]!=0))(population))(populations)
            complexities = jnp.maximum(complexities, 5*self.num_trees*jnp.ones_like(complexities))
            fitness = jnp.concatenate([fitness, complexities[:,:,None]], axis=-1)

        new_populations = evolve_populations(self.jit_evolve_population, 
                                             populations, 
                                             fitness,
                                             key, 
                                             self.current_generation, 
                                             self.migration_period, 
                                             self.migration_size, 
                                             self.reproduction_type_probabilities, 
                                             self.reproduction_probabilities)
        
        self.current_generation += 1
        if self.constant_optimization and self.current_generation >= self.start_constant_optimization:
            self.constant_step_size = jnp.maximum(self.constant_step_size * self.constant_step_size_decay, 0.001) #Update step size for constant optimization
        return self.jit_simplify_constants(new_populations)
    
    def mutate_pair(self, parent1: Array, parent2: Array, keys: Array, reproduction_probability: float) -> Tuple[Array, Array]:
        """
        Mutates a pair of candidates.

        Parameters
        ----------
        parent1 : Array
            First parent candidate.
        parent2 : Array
            Second parent candidate.
        keys : Array
            Random keys.
        reproduction_probability : float
            Probability of a tree to be mutated.

        Returns
        -------
        Tuple[Array, Array]
            Pair of candidates after mutation.
        """
        offspring = jax.vmap(self.mutate_trees, in_axes=[0,1,None,None])(jnp.stack([parent1, parent2]), keys, reproduction_probability, self.variable_array)
        return offspring[0], offspring[1]

    def sample_pair(self, parent1: Array, parent2: Array, keys: Array, reproduction_probability: float) -> Tuple[Array, Array]:
        """
        Samples a pair of candidates.

        Parameters
        ----------
        parent1 : Array
            First parent candidate.
        parent2 : Array
            Second parent candidate.
        keys : Array
            Random keys.
        reproduction_probability : float
            Probability of a tree to be sampled.

        Returns
        -------
        Tuple[Array, Array]
            Pair of candidates.
        """
        offspring = jax.vmap(lambda _keys: jax.vmap(self.sample_tree, in_axes=[0, None, 0])(_keys, self.max_init_depth, self.variable_array), in_axes=[1])(keys)
        return offspring[0], offspring[1]

    def simplify_constants_in_row(self, i: int, carry: Tuple[Array, Array, Array]) -> Tuple[Array, Array, Array]:
        """
        Simplifies the constants in a tree.

        Parameters
        ----------
        i : int
            Index of the node.
        carry : Tuple[Array, Array, Array]
            Tuple containing the tree, tree indices, and empty tree.

        Returns
        -------
        Tuple[Array, Array, Array]
            Simplified tree.
        """
        tree, tree_indices, empty_tree = carry

        last_node_idx = jnp.sum(tree[:,0]==0)
        f_idx, a_idx, b_idx, constant = tree[i]

        evaluated_subtree = tree.at[i].set(jnp.array([1.0, -1.0, -1.0, jax.lax.switch(f_idx.astype(int), self.node_function_list, tree[a_idx.astype(int), -1], tree[b_idx.astype(int), -1], jnp.zeros(1))]))
        
        one_branch_tree = jnp.where((tree_indices < i) & (tree_indices >= last_node_idx + 1), jnp.roll(tree, 1, axis=0), evaluated_subtree)
        one_branch_tree = jnp.where(tree_indices < last_node_idx + 1, empty_tree, one_branch_tree)
        one_branch_tree = one_branch_tree.at[:,1:3].set(jnp.where((one_branch_tree[:,1:3] < a_idx) & (one_branch_tree[:,1:3] > -1), one_branch_tree[:,1:3] + 1, one_branch_tree[:,1:3]))
        
        two_branch_tree = jnp.where((tree_indices < i) & (tree_indices >= last_node_idx + 2), jnp.roll(tree, 2, axis=0), evaluated_subtree)
        two_branch_tree = jnp.where(tree_indices < last_node_idx + 2, empty_tree, two_branch_tree)
        two_branch_tree = two_branch_tree.at[:,1:3].set(jnp.where((two_branch_tree[:,1:3] < b_idx) & (two_branch_tree[:,1:3] > -1), two_branch_tree[:,1:3] + 2, two_branch_tree[:,1:3]))
        
        new_tree = jax.lax.select((tree[a_idx.astype(int), 0] == 1) & (b_idx == -1), one_branch_tree, tree)
        new_tree = jax.lax.select((tree[a_idx.astype(int), 0] == 1) & (tree[b_idx.astype(int), 0] == 1), two_branch_tree, new_tree)

        new_tree = jax.lax.select(a_idx > -1, new_tree, tree)

        return (new_tree, tree_indices, empty_tree)

    def simplify_constants(self, tree: Array) -> Array:
        """
        Simplifies a tree by evaluating constant subtrees.

        Parameters
        ----------
        tree : Array
            Tree to be simplified.

        Returns
        -------
        Array
            Simplified tree.
        """
        tree, _, _ = jax.lax.fori_loop(0, self.max_nodes, self.simplify_constants_in_row, (tree, self.tree_indices, self.empty_tree))

        return tree

    def evaluate_row_from_tree(self, i: int, carry: Tuple[Array, Array], node_function_list: list) -> Tuple[Array, Array]:
        """
        Evaluates a node given inputs.

        Parameters
        ----------
        i : int
            Index of the node.
        carry : Tuple[Array, Array]
            Tuple containing the tree and data.
        node_function_list : list
            List of functions corresponding to nodes.

        Returns
        -------
        Tuple[Array, Array]
            Evaluated node.
        """

        tree, data = carry
        f_idx, a_idx, b_idx, constant = tree[i]  # Get node function, index of first and second operand, and constant value of node (which will be 0 if the node function is not 1)

        x = tree[a_idx.astype(int), 3]  # Value of first operand
        y = tree[b_idx.astype(int), 3]  # Value of second operand
        value = jax.lax.select(f_idx == 1, constant, jax.lax.switch(f_idx.astype(int), node_function_list, x, y, data))  # Computes value of the node

        tree = tree.at[i, 3].set(value)  # Store value

        return (tree, data)

    def iterate_through_tree(self, tree: Array, data: Array) -> Array:
        """
        Loops through a tree to compute the value of each node bottom up.

        Parameters
        ----------
        tree : Array
            Tree to be evaluated.
        data : Array
            Data for evaluation.

        Returns
        -------
        Array
            Value of the root node.
        """

        x, _ = jax.lax.fori_loop(0, self.max_nodes, self.jit_evaluate_row_from_tree, (tree, data)) #Iterate over the tree to compute the value of each node
        return x[-1, -1] #Return the value of the root node

    def tree_evaluator(self, candidate: Array, data: Array) -> Array:
        """
        Calls the evaluation function for each tree in a candidate.

        Parameters
        ----------
        candidate : Array
            Candidate to be evaluated.
        data : Array
            Data for evaluation.

        Returns
        -------
        Array
            Result of each tree.
        """

        # data = jnp.atleast_1d(data)
        result = jax.vmap(self.iterate_through_tree, in_axes=[0, None])(candidate, data) #Evaluate each tree in the candidate
        # return jnp.squeeze(result)
        return result

    def evaluate_population(self, populations: Array, data: Tuple, key: PRNGKey) -> Tuple[Array, Array]:
        """
        Evaluates every candidate in population and assigns a fitness. Optionally, the constants in the candidates are optimized.

        Parameters
        ----------
        populations : Array
            Population of candidates.
        data : Tuple
            The data required to evaluate the population.
        key : PRNGKey
            Random key.

        Returns
        -------
        Tuple[Array, Array]
            Fitness and evaluated or optimized population.
        """

        # Flatten the populations so they can be distributed over the devices
        flat_populations = populations.reshape(self.num_populations * self.population_size, *populations.shape[2:])
        data = jax.device_put(data, self.data_mesh)

        fitness = self.jit_eval(flat_populations, data)  # Evaluate the candidates

        # Optimize constants of the best candidates in the current generation
        if self.constant_optimization and self.current_generation >= self.start_constant_optimization:
            self.optimizer = self.optimizer_class(self.constant_step_size)

            # Get best candidates
            best_candidates_idx = jnp.argsort(fitness)[:self.optimize_constants_elite]
            best_candidates = flat_populations[best_candidates_idx]

            # Optimize constants of the best candidates
            optimized_fitness, optimized_population = self.jit_optimize(best_candidates, data, jr.split(key, self.optimize_constants_elite), self.constant_step_size)

            optimized_population = jax.block_until_ready(optimized_population)

            # Store updated candidates and fitness
            flat_populations = flat_populations.at[best_candidates_idx].set(optimized_population)
            fitness = fitness.at[best_candidates_idx].set(optimized_fitness)

        self.update_pareto_front(fitness, flat_populations)

        # Reshape the populations into the subpopulations
        fitness = fitness.reshape((self.num_populations, self.population_size))
        populations = flat_populations.reshape((self.num_populations, self.population_size, *flat_populations.shape[1:]))

        return fitness, populations
    
    def optimize_constants_epoch(self, carry: Tuple[Array, Array, Tuple], x: int) -> Tuple[Tuple[Array, Array, Tuple], Tuple[Array, Array]]:
        """
        Applies one step of constant optimization to a batch of candidates.

        Parameters
        ----------
        carry : Tuple[Array, Array, Tuple]
            Tuple containing candidates, states, and data.
        x : int
            Unused parameter for scan.

        Returns
        -------
        Tuple[Tuple[Array, Array, Tuple], Tuple[Array, Array]]
            Tuple containing updated candidates, states, and data, and the original candidates and loss.
        """
        candidates, states, data = carry
        loss, gradients = self.vmap_gradients(candidates[..., 3:], candidates[..., :3], data)  # Parallel computation of the loss and gradients

        # Regularize invalid solutions
        nan_or_inf = jax.vmap(lambda f: jnp.isinf(f) + jnp.isnan(f))(loss)
        loss = jnp.where(nan_or_inf, jnp.ones(loss.shape) * self.max_fitness, loss)

        loss = jnp.minimum(loss, self.max_fitness * jnp.ones_like(loss))

        updates, states = jax.vmap(self.optimizer.update)(gradients, states, candidates[..., 3])  # Parallel computation of the updates
        new_candidates = candidates.at[..., 3:].set(jax.vmap(lambda t, u: t + u)(candidates[..., 3:], updates))  # Parallel updating of constants

        return (new_candidates, states, data), (candidates, loss)

    def optimize_constants_with_gradients(self, candidates: Array, data: Tuple, key: PRNGKey, step_size: float, n_epoch: int) -> Tuple[Array, Array]:
        """
        optimizes the constants in the candidates.

        Parameters
        ----------
        candidates : Array
            Candidate solutions.
        data : Tuple
            The data required to evaluate the population.
        key : PRNGKey
            Random key.
        n_epoch : int
            Number of steps to optimize constants.

        Returns
        -------
        Tuple[Array, Array]
            Optimized and evaluated candidate.
        """

        states = jax.vmap(self.optimizer.init)(candidates[..., 3:])  # Initialize optimizers for each candidate

        _, out = jax.lax.scan(self.optimize_constants_epoch, (candidates, states, data), length=n_epoch)

        new_candidates, loss = out

        fitness = jnp.min(loss, axis=0)  # Get best fitness during constant optimization
        candidates = jax.vmap(lambda t, i: t[i], in_axes=[1, 0])(new_candidates, jnp.argmin(loss, axis=0))  # Get best candidate during constant optimization

        return fitness, candidates

    def optimize_constants_generation(self, carry: Tuple[Array, Tuple, PRNGKey], x: int) -> Tuple[Tuple[Array, Tuple, PRNGKey], float]:
        """
        optimizes a generation of candidates.

        Parameters
        ----------
        carry : Tuple[Array, Tuple, PRNGKey]
            Tuple containing candidate, data, and key.
        x : int
            Unused parameter for scan.

        Returns
        -------
        Tuple[Tuple[Array, Tuple, PRNGKey], float]
            Tuple containing updated candidate, data, and key, and the best fitness.
        """

        candidate, data, key, step_size = carry

        key, sample_key = jr.split(key)

        mask = candidate[..., 0] == 1.0 #Only samples mutations for the nodes that contain a constant
        mutations = jax.vmap(lambda _key: step_size * jr.normal(_key, shape=(self.num_trees, self.max_nodes,)) * mask)(jr.split(sample_key, self.n_offspring))
        mutations = jnp.vstack([jnp.zeros((1, self.num_trees, self.max_nodes)), mutations])

        offspring = jax.vmap(lambda m: candidate.at[..., 3].set(candidate[..., 3] + m))(mutations)

        fitness = self.vmap_trees(offspring[..., 3:], offspring[..., :3], data)

        # Regularize invalid solutions
        nan_or_inf = jax.vmap(lambda f: jnp.isinf(f) + jnp.isnan(f))(fitness)
        fitness = jnp.where(nan_or_inf, jnp.ones(self.n_offspring + 1) * self.max_fitness, fitness)

        fitness = jnp.minimum(fitness, self.max_fitness * jnp.ones_like(fitness))

        return (offspring[jnp.argmin(fitness)], data, key, step_size), jnp.min(fitness)

    def optimize_constants_with_evolution(self, candidate: Array, data: Tuple, key: PRNGKey, step_size: float, n_iterations: int) -> Tuple[float, Array]:
        """
        optimizes a candidate using Evolution Strategies (ES).

        Parameters
        ----------
        candidate : Array
            Candidate solution.
        data : Tuple
            The data required to evaluate the population.
        key : PRNGKey
            Random key.
        n_iterations : int
            Number of iterations for optimization.

        Returns
        -------
        Tuple[float, Array]
            Best fitness and optimized candidate.
        """

        (new_candidate, _, _, _), fitness = jax.lax.scan(self.optimize_constants_generation, (candidate, data, key, step_size), length=n_iterations)

        return jnp.min(fitness), new_candidate

    def punish_duplicates(self, population: Array, fitness: Array) -> Tuple[Array, Array]:
        """
        Punishes duplicate candidates by setting their fitness to the maximum fitness.

        Parameters
        ----------
        population : Array
            Population of candidates.
        fitness : Array
            Fitness of candidates.

        Returns
        -------
        Tuple[Array, Array]
            Population and fitness with duplicates punished.
        """

        _, indices, counts = jnp.unique(population, return_index=True, return_counts=True, axis=0, size=self.population_size)
        population = population[indices]
        fitness = fitness[indices]
        return population, jnp.where(counts > 0, fitness, self.max_fitness)

    def update_pareto_front(self, fitness: Array, population: Array) -> None:
        """
        Updates the Pareto front with the current population.

        Parameters
        ----------
        current_fitness : Array
            Fitness of the current population.
        current_population : Array
            Current population.
        """
        current_pareto_fitness, current_pareto_solutions = self.pareto_front

        _fitness = jnp.concatenate([fitness, current_pareto_fitness], axis=0)
        _population = jnp.concatenate([population, current_pareto_solutions], axis=0)

        # Compute complexity of the current population
        complexity = jax.vmap(lambda array: jnp.sum(array[:, :, 0] != 0))(_population)[:,None]
        metrics = jnp.concatenate([_fitness[:,None], complexity], axis=-1)

        # For each solution i, check if it's dominated by any other solution j
        
        # Reshape metrics for broadcasting. Shape: (n_solutions, 1, n_objectives)
        metrics_i = jnp.expand_dims(metrics, axis=1)
        
        # Shape: (1, n_solutions, n_objectives)
        metrics_j = jnp.expand_dims(metrics, axis=0)
        
        # Check if j dominates i: 
        # 1. j is better or equal in all objectives
        # 2. j is strictly better in at least one objective
        j_better_or_equal = (metrics_j <= metrics_i)
        j_strictly_better = (metrics_j < metrics_i)
        
        # For each pair (i,j), check if j dominates i
        j_dominates_i = jnp.all(j_better_or_equal, axis=2) & jnp.any(j_strictly_better, axis=2)
        
        # Don't compare solutions with themselves
        mask = ~jnp.eye(metrics.shape[0], dtype=bool)
        j_dominates_i = j_dominates_i & mask
        
        # A solution is non-dominated if it's not dominated by any other solution
        dominated_by_others = ~jnp.any(j_dominates_i, axis=1)
        
        pareto_indices = jnp.nonzero(dominated_by_others)[0]

        pareto_solutions, unique_indices = jnp.unique(_population[pareto_indices], return_index=True, axis=0)

        self.pareto_front = (_fitness[pareto_indices][unique_indices], pareto_solutions)

    def print_pareto_front(self, save: bool = False, path_to_file: str = None) -> None:
        """
        Prints the Pareto front.

        Parameters
        ----------
        save : bool, optional
            Whether to save the Pareto front to a file.
        path_to_file : str, optional
            Name of the file to save the Pareto front.
        """

        # Check if path exists when saving is requested
        if save:
            assert path_to_file is not None, "A file name must be provided when saving the Pareto front"
            import os
            # Check if directory exists and create it if needed
            directory = os.path.dirname(path_to_file)
            if directory and not os.path.exists(directory):
                os.makedirs(directory)

        pareto_fitness, pareto_solutions = self.pareto_front

        complexities = jax.vmap(lambda array: jnp.sum(array[:, :, 0] != 0))(pareto_solutions)

        sorted_indices = jnp.argsort(complexities)
        pareto_fitness = pareto_fitness[sorted_indices]
        pareto_solutions = pareto_solutions[sorted_indices]
        complexities = complexities[sorted_indices]

        if save:
            pareto_table = ['Complexity', 'Fitness']
            for i in range(self.num_trees):
                pareto_table.append(f'Equation {i}')
            pareto_table = [tuple(pareto_table)]

        expression_strings = []

        for c in range(complexities.shape[0]):
            string_equations = self.expression_to_string(pareto_solutions[c])
            if string_equations not in expression_strings:
                expression_strings.append(string_equations)
                if self.num_trees>1:
                    print(f"Complexity: {complexities[c]}, fitness: {pareto_fitness[c]}, equations: {string_equations}")
                else:
                    print(f"Complexity: {complexities[c]}, fitness: {pareto_fitness[c]}, equation: {string_equations}")

                if save:
                    if self.num_trees > 1:
                        temp = (complexities[c], pareto_fitness[c])
                        for tree in string_equations:
                            temp += (tree,)
                        pareto_table.append(temp)
                    else:
                        pareto_table.append((complexities[c], pareto_fitness[c], string_equations))

        if save:
            np.savetxt(f'{path_to_file}/pareto_front.csv', pareto_table, delimiter=',', fmt='%s')
    
    def tree_to_string(self, tree: Array) -> str:
        """
        Maps tree to string.

        Parameters
        ----------
        tree : Array
            Tree to be converted to string.

        Returns
        -------
        str
            String representation of tree.
        """

        if tree[-1, 0] == 1:  # constant
            return str(tree[-1, 3])
        elif tree[-1, 1] < 0:  # Variable
            return self.node_to_string[tree[-1, 0].astype(int).item()]
        elif tree[-1, 2] < 0:  # Operator with one operand
            substring = self.tree_to_string(tree[:tree[-1, 1].astype(int) + 1])
            operator_string = self.node_to_string[tree[-1, 0].astype(int).item()]

            if operator_string[0].isalpha() or operator_string[0].isdigit():
                return f"{operator_string}({substring})"
            else:
                return f"({substring}){operator_string}"
            
        else:  # Operator with two operands
            substring1 = self.tree_to_string(tree[:tree[-1, 1].astype(int) + 1])
            substring2 = self.tree_to_string(tree[:tree[-1, 2].astype(int) + 1])
            operator_string = self.node_to_string[tree[-1, 0].astype(int).item()]

            if operator_string in ["+", "-", "*", "/", "**"]:
                return f"({substring1}){self.node_to_string[tree[-1, 0].astype(int).item()]}({substring2})"
            else:
                return f"{operator_string}({substring1}, {substring2})"

    def expression_to_string(self, candidate: Array) -> str:
        """
        Maps trees in a candidate to string.

        Parameters
        ----------
        candidate : Array
            Candidate to be converted to string.

        Returns
        -------
        str
            String representation of candidate.
        """
        if self.num_trees == 1:
            simplified_expression = sympy.parsing.sympy_parser.parse_expr(self.tree_to_string(candidate[0]))

            rounded_expression = simplified_expression

            for a in sympy.preorder_traversal(simplified_expression):
                if isinstance(a, sympy.Float):
                    rounded_expression = rounded_expression.subs(a, sympy.Float(a, 3))

            return rounded_expression

        string_output = []
        for tree in candidate:
            simplified_expression = sympy.parsing.sympy_parser.parse_expr(self.tree_to_string(tree))

            rounded_expression = simplified_expression

            for a in sympy.preorder_traversal(simplified_expression):
                if isinstance(a, sympy.Float):
                    rounded_expression = rounded_expression.subs(a, sympy.Float(a, 3))

            string_output.append(rounded_expression)

        return string_output


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
from kozax.environments.SR_environments.time_series_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class LorenzAttractor(EnvironmentBase):
    """
    Lorenz Attractor environment for symbolic regression tasks.

    Parameters
    ----------
    process_noise : float, optional
        Standard deviation of the process noise. Default is 0.

    Attributes
    ----------
    init_mu : :class:`jax.Array`
        Mean of the initial state distribution.
    init_sd : float
        Standard deviation of the initial state distribution.
    sigma : float
        Parameter sigma of the Lorenz system.
    rho : float
        Parameter rho of the Lorenz system.
    beta : float
        Parameter beta of the Lorenz system.
    V : :class:`jax.Array`
        Process noise covariance matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    """

    def __init__(self, process_noise: float = 0) -> None:
        n_var = 3
        super().__init__(n_var, process_noise)

        self.init_mu = jnp.array([1, 1, 1])
        self.init_sd = 0.1

        self.sigma = 10
        self.rho = 28
        self.beta = 8 / 3
        self.V = self.process_noise * jnp.eye(self.n_var)

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Array:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        :class:`jax.Array`
            Initial states.
        """
        return self.init_mu + self.init_sd * jrandom.normal(key, shape=(batch_size, self.n_var))

    def drift(self, t: float, state: Array, args: Tuple = jnp.array([0.0])) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        u = args[0]
        return jnp.array([self.sigma * (state[1] - state[0]), state[0] * (self.rho - state[2]) - state[1], state[0] * state[1] - self.beta * state[2]])

    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V

class Lorenz96(EnvironmentBase):
    """
    Lorenz 96 environment for symbolic regression tasks.

    Parameters
    ----------
    n_dim : int, optional
        Number of dimensions. Default is 3.
    process_noise : float, optional
        Standard deviation of the process noise. Default is 0.

    Attributes
    ----------
    F : float
        Forcing term.
    init_mu : :class:`jax.Array`
        Mean of the initial state distribution.
    init_sd : float
        Standard deviation of the initial state distribution.
    V : :class:`jax.Array`
        Process noise covariance matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    """

    def __init__(self, n_dim: int = 3, process_noise: float = 0) -> None:
        n_var = n_dim
        super().__init__(n_var, process_noise)

        self.F = 8
        self.init_mu = jnp.ones(self.n_var) * self.F
        self.init_sd = 0.1

        self.V = self.process_noise * jnp.eye(self.n_var)

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Array:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        :class:`jax.Array`
            Initial states.
        """
        return self.init_mu + self.init_sd * jrandom.normal(key, shape=(batch_size, self.n_var))

    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        f = lambda x_cur, x_next, x_prev1, x_prev2: (x_next - x_prev2) * x_prev1 - x_cur + self.F
        return jax.vmap(f)(state, jnp.roll(state, -1), jnp.roll(state, 1), jnp.roll(state, 2))

    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
from kozax.environments.control_environments.control_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class CartPole(EnvironmentBase):
    """
    CartPole environment for control tasks.

    Parameters
    ----------
    process_noise : float
        Standard deviation of the process noise.
    obs_noise : float
        Standard deviation of the observation noise.
    n_obs : int, optional
        Number of observations. Default is 4.

    Attributes
    ----------
    n_var : int
        Number of variables in the state.
    n_control_inputs : int
        Number of control inputs.
    n_targets : int
        Number of targets.
    n_dim : int
        Number of dimensions.
    init_bounds : :class:`jax.Array`
        Bounds for initial state sampling.
    Q : :class:`jax.Array`
        Process noise covariance matrix.
    R : :class:`jax.Array`
        Observation noise covariance matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    sample_params(batch_size, mode, ts, key)
        Samples parameters for the environment.
    initialize_parameters(params, ts)
        Initializes the parameters of the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    fitness_function(state, control, target, ts)
        Computes the fitness function for the environment.
    terminate_event(state, **kwargs)
        Checks if the termination condition is met.
    """

    def __init__(self, process_noise: float = 0.0, obs_noise: float = 0.0, n_obs: int = 4) -> None:
        self.n_var = 4
        self.n_control_inputs = 1
        self.n_targets = 0
        self.n_dim = 1
        self.init_bounds = jnp.array([0.05, 0.05, 0.05, 0.05])
        super().__init__(process_noise, obs_noise, self.n_var, self.n_control_inputs, self.n_dim, n_obs)

        self.Q = jnp.array(0)
        self.R = jnp.array([[0.0]])

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Tuple[Array, Array]:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        x0 : :class:`jax.Array`
            Initial states.
        targets : :class:`jax.Array`
            Target states.
        """
        init_key, target_key = jrandom.split(key)
        x0 = jrandom.uniform(init_key, shape=(batch_size, self.n_var), minval=-self.init_bounds, maxval=self.init_bounds)
        targets = jnp.zeros((batch_size, self.n_targets))
        return x0, targets
    
    def sample_params(self, batch_size: int, mode: str, ts: Array, key: jrandom.PRNGKey) -> Array:
        """
        Samples parameters for the environment.

        Parameters
        ----------
        batch_size : int
            Number of parameters to sample.
        mode : str
            Mode for sampling parameters.
        ts : :class:`jax.Array`
            Time steps.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        :class:`jax.Array`
            Sampled parameters.
        """
        return jnp.zeros((batch_size))

    def initialize_parameters(self, params: Array, ts: Array) -> None:
        """
        Initializes the parameters of the environment.

        Parameters
        ----------
        params : :class:`jax.Array`
            Parameters to initialize.
        ts : :class:`jax.Array`
            Time steps.
        """
        _ = params
        self.g = 9.81
        self.pole_mass = 0.1
        self.pole_length = 0.5
        self.cart_mass = 1
        
        self.G = jnp.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]])
        self.V = self.process_noise * self.G

        self.C = jnp.eye(self.n_var)[:self.n_obs]
        self.W = self.obs_noise * jnp.eye(self.n_obs)

    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        control = jnp.squeeze(args)
        control = jnp.clip(control, -1, 1)
        x, theta, x_dot, theta_dot = state

        cos_theta = jnp.cos(theta)
        sin_theta = jnp.sin(theta)
        
        theta_acc = (self.g * sin_theta - cos_theta * (
            control + self.pole_mass * self.pole_length * theta_dot**2 * sin_theta
            ) / (self.cart_mass + self.pole_mass)) / (
                self.pole_length * (4/3 - (self.pole_mass * cos_theta**2) / (self.cart_mass + self.pole_mass)))

        x_acc = (control + self.pole_mass * self.pole_length * (theta_dot**2 * sin_theta - theta_acc * cos_theta)) / (self.cart_mass + self.pole_mass)

        return jnp.array([
            x_dot,
            theta_dot,
            x_acc,
            theta_acc
        ])
    
    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V
    
    def fitness_function(self, state: Array, control: Array, target: Array, ts: Array) -> float:
        """
        Computes the fitness function for the environment.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        control : :class:`jax.Array`
            Control inputs.
        target : :class:`jax.Array`
            Target states.
        ts : :class:`jax.Array`
            Time steps.

        Returns
        -------
        float
            Fitness value.
        """
        invalid_points = jax.vmap(lambda _x, _u: jnp.any(jnp.isinf(_x)) + jnp.isnan(_u))(state, control[:, 0])
        punishment = jnp.ones_like(invalid_points)

        costs = jnp.where(invalid_points, punishment, jnp.zeros_like(punishment))

        return jnp.sum(costs)
    
    def terminate_event(self, state: Array, **kwargs) -> bool:
        """
        Checks if the termination condition is met.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        kwargs : dict
            Additional arguments.

        Returns
        -------
        bool
            True if the termination condition is met, False otherwise.
        """
        return (jnp.abs(state[1]) > 0.2) | (jnp.abs(state[0]) > 4.8) | jnp.any(jnp.isnan(state)) | jnp.any(jnp.isinf(state))


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jr
from typing import Tuple
from jaxtyping import Array
from jax.random import PRNGKey

def sample_indices(carry: Tuple[PRNGKey, Array, float]) -> Tuple[PRNGKey, Array, float]:
    """
    Samples indices of the trees in a candidate that will be mutated.

    Parameters
    ----------
    carry : tuple of (PRNGKey, Array, float)
        Tuple containing the random key, indices of trees, and reproduction probability.

    Returns
    -------
    tuple of (PRNGKey, Array, float)
        Updated tuple with the random key, indices of trees to be mutated, and reproduction probability.
    """
    key, indices, reproduction_probability = carry
    indices = jr.bernoulli(key, p=reproduction_probability, shape=indices.shape) * 1.0
    return (jr.split(key, 1)[0], indices, reproduction_probability)

def find_end_idx(carry: Tuple[Array, int, int]) -> Tuple[Array, int, int]:
    """
    Finds the index of the last node in a subtree.

    Parameters
    ----------
    carry : tuple of (Array, int, int)
        Tuple containing the tree, the number of open slots, and the current node index.

    Returns
    -------
    tuple of (Array, int, int)
        Updated tuple with the tree, open slots, and current node index.
    """
    tree, open_slots, counter = carry
    _, idx1, idx2, _ = tree[counter]
    open_slots -= 1  # Reduce open slot for current node
    open_slots = jax.lax.select(idx1 < 0, open_slots, open_slots + 1)  # Increase the open slots for a child
    open_slots = jax.lax.select(idx2 < 0, open_slots, open_slots + 1)  # Increase the open slots for a child
    counter -= 1
    return (tree, open_slots, counter)

def check_invalid_cx_nodes(carry: Tuple[Array, Array, Array, int, int, Array, Array]) -> bool:
    """
    Checks if the sampled subtrees are different and if the trees after crossover are valid.

    Parameters
    ----------
    carry : tuple of (Array, Array, Array, int, int, Array, Array)
        Tuple containing the trees, node indices, and other parameters.

    Returns
    -------
    bool
        If the sampled nodes are valid nodes for crossover.
    """
    tree1, tree2, _, node_idx1, node_idx2, _, _ = carry

    _, _, end_idx1 = jax.lax.while_loop(lambda carry: carry[1] > 0, find_end_idx, (tree1, 1, node_idx1))
    _, _, end_idx2 = jax.lax.while_loop(lambda carry: carry[1] > 0, find_end_idx, (tree2, 1, node_idx2))

    subtree_size1 = node_idx1 - end_idx1
    subtree_size2 = node_idx2 - end_idx2

    empty_nodes1 = jnp.sum(tree1[:, 0] == 0)
    empty_nodes2 = jnp.sum(tree2[:, 0] == 0)

    # Check if the subtrees can be inserted
    return (empty_nodes1 < subtree_size2 - subtree_size1) | (empty_nodes2 < subtree_size1 - subtree_size2)

def sample_cx_nodes(carry: Tuple[Array, Array, Array, int, int, Array, Array]) -> Tuple[Array, Array, Array, int, int, Array, Array]:
    """
    Samples nodes in a pair of trees for crossover.

    Parameters
    ----------
    carry : tuple of (Array, Array, Array, int, int, Array, Array)
        Tuple containing the trees, node indices, and other parameters.

    Returns
    -------
    tuple of (Array, Array, Array, int, int, Array, Array)
        Updated tuple with the sampled nodes.
    """
    tree1, tree2, keys, _, _, node_ids, operator_indices = carry
    key1, key2 = keys

    # Sample nodes from the non-empty nodes, with higher probability for operator nodes
    cx_prob1 = jnp.isin(tree1[:, 0], operator_indices)
    cx_prob1 = jnp.where(tree1[:, 0] == 0, cx_prob1, cx_prob1 + 1)
    node_idx1 = jr.choice(key1, node_ids, p=cx_prob1 * 1.0)

    cx_prob2 = jnp.isin(tree2[:, 0], operator_indices)
    cx_prob2 = jnp.where(tree2[:, 0] == 0, cx_prob2, cx_prob2 + 1)
    node_idx2 = jr.choice(key2, node_ids, p=cx_prob2 * 1.0)

    return (tree1, tree2, jr.split(key1), node_idx1, node_idx2, node_ids, operator_indices)

def tree_crossover(tree1: Array, 
                   tree2: Array, 
                   keys: Array,
                   node_ids: Array,
                   operator_indices: Array) -> Tuple[Array, Array]:
    """
    Applies crossover to a pair of trees to produce two new trees.

    Parameters
    ----------
    tree1 : Array
        First tree.
    tree2 : Array
        Second tree.
    keys : Array
        Random keys.
    node_ids : Array
        Indices of all the nodes in the trees.
    operator_indices : Array
        The indices that belong to operator nodes.

    Returns
    -------
    tuple of (Array, Array)
        Pair of new trees.
    """
    # Define indices of the nodes
    tree_indices = jnp.tile(node_ids[:, None], reps=(1, 4))
    key1, key2 = keys

    # Define last node in tree
    last_node_idx1 = jnp.sum(tree1[:, 0] == 0)
    last_node_idx2 = jnp.sum(tree2[:, 0] == 0)

    # Randomly select nodes for crossover
    _, _, _, node_idx1, node_idx2, _, _ = sample_cx_nodes((tree1, tree2, jr.split(key1), 0, 0, node_ids, operator_indices))

    # Reselect until valid crossover nodes have been found
    _, _, _, node_idx1, node_idx2, _, _ = jax.lax.while_loop(check_invalid_cx_nodes, sample_cx_nodes, (tree1, tree2, jr.split(key2), node_idx1, node_idx2, node_ids, operator_indices))

    # Retrieve subtrees of selected nodes
    _, _, end_idx1 = jax.lax.while_loop(lambda carry: carry[1] > 0, find_end_idx, (tree1, 1, node_idx1))
    _, _, end_idx2 = jax.lax.while_loop(lambda carry: carry[1] > 0, find_end_idx, (tree2, 1, node_idx2))

    # Initialize children
    child1 = jnp.tile(jnp.array([0.0, -1.0, -1.0, 0.0]), (len(node_ids), 1))
    child2 = jnp.tile(jnp.array([0.0, -1.0, -1.0, 0.0]), (len(node_ids), 1))

    # Compute subtree sizes
    subtree_size1 = node_idx1 - end_idx1
    subtree_size2 = node_idx2 - end_idx2

    # Insert nodes before subtree in children
    child1 = jnp.where(tree_indices >= node_idx1 + 1, tree1, child1)
    child2 = jnp.where(tree_indices >= node_idx2 + 1, tree2, child2)
    
    # Align nodes after subtree with first open spot after new subtree in children
    rolled_tree1 = jnp.roll(tree1, subtree_size1 - subtree_size2, axis=0)
    rolled_tree2 = jnp.roll(tree2, subtree_size2 - subtree_size1, axis=0)

    # Insert nodes after subtree in children
    child1 = jnp.where((tree_indices >= node_idx1 - subtree_size2 - (end_idx1 - last_node_idx1)) & (tree_indices < node_idx1 + 1 - subtree_size2), rolled_tree1, child1)
    child2 = jnp.where((tree_indices >= node_idx2 - subtree_size1 - (end_idx2 - last_node_idx2)) & (tree_indices < node_idx2 + 1 - subtree_size1), rolled_tree2, child2)

    # Update index references to moved nodes in staying nodes
    child1 = child1.at[:, 1:3].set(jnp.where((child1[:, 1:3] < (node_idx1 - subtree_size1 + 1)) & (child1[:, 1:3] > -1), child1[:, 1:3] + (subtree_size1 - subtree_size2), child1[:, 1:3]))
    child2 = child2.at[:, 1:3].set(jnp.where((child2[:, 1:3] < (node_idx2 - subtree_size2 + 1)) & (child2[:, 1:3] > -1), child2[:, 1:3] + (subtree_size2 - subtree_size1), child2[:, 1:3]))

    # Align subtree with the selected node in children
    rolled_subtree1 = jnp.roll(tree1, node_idx2 - node_idx1, axis=0)
    rolled_subtree2 = jnp.roll(tree2, node_idx1 - node_idx2, axis=0)

    # Update index references in subtree
    rolled_subtree1 = rolled_subtree1.at[:, 1:3].set(jnp.where(rolled_subtree1[:, 1:3] > -1, rolled_subtree1[:, 1:3] + (node_idx2 - node_idx1), -1))
    rolled_subtree2 = rolled_subtree2.at[:, 1:3].set(jnp.where(rolled_subtree2[:, 1:3] > -1, rolled_subtree2[:, 1:3] + (node_idx1 - node_idx2), -1))

    # Insert subtree in selected node in children
    child1 = jnp.where((tree_indices >= node_idx1 + 1 - subtree_size2) & (tree_indices < node_idx1 + 1), rolled_subtree2, child1)
    child2 = jnp.where((tree_indices >= node_idx2 + 1 - subtree_size1) & (tree_indices < node_idx2 + 1), rolled_subtree1, child2)
    
    return child1, child2

def full_crossover(tree1: Array, 
                   tree2: Array, 
                   keys: Array,
                   node_ids: Array,
                   operator_indices: Array) -> Tuple[Array, Array]:
    """
    Swaps the entire trees between two candidates.

    Parameters
    ----------
    tree1 : Array
        First tree.
    tree2 : Array
        Second tree.
    keys : Array
        Random keys.
    node_ids : Array
        Indices of all the nodes in the trees.
    operator_indices : Array
        The indices that belong to operator nodes.

    Returns
    -------
    tuple of (Array, Array)
        Swapped trees.
    """
    return tree2, tree1

def crossover(tree1: Array, 
              tree2: Array, 
              keys: Array,
              node_ids: Array,
              operator_indices: Array,
              crossover_types: int) -> Tuple[Array, Array]:
    """
    Applies crossover to a pair of trees based on the crossover type.

    Parameters
    ----------
    tree1 : Array
        First tree.
    tree2 : Array
        Second tree.
    keys : Array
        Random keys.
    node_ids : Array
        Indices of all the nodes in the trees.
    operator_indices : Array
        The indices that belong to operator nodes.
    crossover_types : int
        Type of crossover to apply.

    Returns
    -------
    tuple of (Array, Array)
        Pair of new trees.
    """
    return jax.lax.cond(crossover_types, tree_crossover, full_crossover, tree1, tree2, keys, node_ids, operator_indices)

def check_different_tree(parent1: Array, parent2: Array, child1: Array, child2: Array) -> bool:
    """
    Checks if the offspring are different from the parents.

    Parameters
    ----------
    parent1 : Array
        First parent tree.
    parent2 : Array
        Second parent tree.
    child1 : Array
        First child tree.
    child2 : Array
        Second child tree.

    Returns
    -------
    bool
        True if the offspring are different from the parents, False otherwise.
    """
    size1 = jnp.sum(child1[:, 0] != 0)
    size2 = jnp.sum(child2[:, 0] != 0)
    
    check1 = (jnp.all(parent1 == child1) | jnp.all(parent2 == child1))
    check2 = (jnp.all(parent1 == child2) | jnp.all(parent2 == child2))

    return ((check1 | check2) & ((size1 > 1) & (size2 > 1))) | (size1 == 0)

def check_different_trees(carry: Tuple[Array, Array, Array, Array, Array, float, Array, Array]) -> bool:
    """
    Checks if the offspring are different from the parents for all trees in the population.

    Parameters
    ----------
    carry : tuple of (Array, Array, Array, Array, Array, float, Array, Array)
        Tuple containing the parent trees, child trees, and other parameters.

    Returns
    -------
    bool
        True if the offspring are different from the parents for all trees, False otherwise.
    """
    parent1, parent2, child1, child2, _, _, _, _ = carry
    return jnp.all(jax.vmap(check_different_tree)(parent1, parent2, child1, child2))

def safe_crossover(carry: Tuple[Array, Array, Array, Array, Array, float, Array, Array]) -> Tuple[Array, Array, Array, Array, Array, float, Array, Array]:
    """
    Ensures that the crossover produces valid offspring.

    Parameters
    ----------
    carry : tuple of (Array, Array, Array, Array, Array, float, Array, Array)
        Tuple containing the parent trees, child trees, and other parameters.

    Returns
    -------
    tuple of (Array, Array, Array, Array, Array, float, Array, Array)
        Updated tuple with the parent trees, child trees, and other parameters.
    """
    parent1, parent2, _, _, keys, reproduction_probability, node_ids, operator_indices = carry
    index_key, type_key, new_key = jr.split(keys[0, 0], 3)
    _, cx_indices, _ = jax.lax.while_loop(lambda carry: jnp.sum(carry[1]) == 0, sample_indices, (index_key, jnp.zeros(parent1.shape[0]), reproduction_probability))
    crossover_types = jr.bernoulli(type_key, p=0.9, shape=(parent1.shape[0],))
    offspring1, offspring2 = jax.vmap(crossover, in_axes=[0, 0, 0, None, None, 0])(parent1, parent2, keys, node_ids, operator_indices, crossover_types)
    child1 = jnp.where(cx_indices[:, None, None] * jnp.ones_like(parent1), offspring1, parent1)
    child2 = jnp.where(cx_indices[:, None, None] * jnp.ones_like(parent2), offspring2, parent2)

    keys = jr.split(new_key, keys.shape[:-1])

    return parent1, parent2, child1, child2, keys, reproduction_probability, node_ids, operator_indices

def crossover_trees(parent1: Array, 
                    parent2: Array, 
                    keys: Array, 
                    reproduction_probability: float, 
                    max_nodes: int, 
                    operator_indices: Array) -> Tuple[Array, Array]:
    """
    Applies crossover to the trees in a pair of candidates.

    Parameters
    ----------
    parent1 : Array
        First parent candidate.
    parent2 : Array
        Second parent candidate.
    keys : Array
        Random keys.
    reproduction_probability : float
        Probability of a tree to be mutated.
    max_nodes : int
        Max number of nodes in a tree.
    operator_indices : Array
        The indices that belong to operator nodes.

    Returns
    -------
    tuple of (Array, Array)
        Pair of candidates after crossover.
    """
    _, _, child1, child2, _, _, _, _ = jax.lax.while_loop(check_different_trees, safe_crossover, (
        parent1, parent2, jnp.zeros_like(parent1), jnp.zeros_like(parent2), keys, reproduction_probability, jnp.arange(max_nodes), operator_indices))
    
    return child1, child2


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
from kozax.environments.SR_environments.time_series_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class LotkaVolterra(EnvironmentBase):
    """
    Lotka-Volterra environment for symbolic regression tasks.

    Parameters
    ----------
    process_noise : float, optional
        Standard deviation of the process noise. Default is 0.

    Attributes
    ----------
    init_mu : :class:`jax.Array`
        Mean of the initial state distribution.
    init_sd : float
        Standard deviation of the initial state distribution.
    alpha : float
        Parameter alpha of the Lotka-Volterra system.
    beta : float
        Parameter beta of the Lotka-Volterra system.
    delta : float
        Parameter delta of the Lotka-Volterra system.
    gamma : float
        Parameter gamma of the Lotka-Volterra system.
    V : :class:`jax.Array`
        Process noise covariance matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    """

    def __init__(self, process_noise: float = 0) -> None:
        n_var = 2
        super().__init__(n_var, process_noise)

        self.init_mu = jnp.array([10, 10])
        self.init_sd = 2

        self.alpha = 1.1
        self.beta = 0.4
        self.delta = 0.1
        self.gamma = 0.4
        self.V = self.process_noise * jnp.eye(self.n_var)

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Array:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        :class:`jax.Array`
            Initial states.
        """
        return jrandom.uniform(key, shape=(batch_size, self.n_var), minval=5, maxval=15)
    
    def drift(self, t: float, state: Array, args: float = jnp.array([0.0])) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        control = args[0]
        return jnp.array([self.alpha * state[0] - self.beta * state[0] * state[1], self.delta * state[0] * state[1] - self.gamma * state[1] + control])

    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

from abc import ABC, abstractmethod
from typing import Tuple, Callable
from jaxtyping import Array

class BaseFitnessFunction(ABC):
    """
    Abstract base class for evaluating candidates in genetic programming.

    Methods
    -------
    __call__(candidate, data, tree_evaluator)
        Evaluates the candidate on a task.
    """

    @abstractmethod
    def __call__(self, candidate: Array, data: Tuple, tree_evaluator: Callable) -> float:
        """
        Evaluates the candidate on a task.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            The candidate solution to be evaluated.
        data : :class:`tuple`
            The data required to evaluate the candidate.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        float
            Fitness of the candidate.
        """
        raise NotImplementedError


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
import jax.numpy as jnp
import jax.random as jrandom
from kozax.environments.control_environments.control_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class Acrobot(EnvironmentBase):
    """
    Acrobot environment for control tasks.

    Parameters
    ----------
    process_noise : float
        Standard deviation of the process noise.
    obs_noise : float
        Standard deviation of the observation noise.
    n_obs : int, optional
        Number of observations. Default is 4.

    Attributes
    ----------
    n_var : int
        Number of variables in the state.
    n_control_inputs : int
        Number of control inputs.
    n_targets : int
        Number of targets.
    n_dim : int
        Number of dimensions.
    init_bounds : :class:`jax.Array`
        Bounds for initial state sampling.
    R : :class:`jax.Array`
        Control cost matrix.

    Methods
    -------
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    sample_params(batch_size, mode, ts, key)
        Samples parameters for the environment.
    f_obs(key, t_x)
        Computes the observation function.
    initialize_parameters(params, ts)
        Initializes the parameters of the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    fitness_function(state, control, target, ts)
        Computes the fitness function for the environment.
    cond_fn_nan(t, y, args, **kwargs)
        Checks for NaN or infinite values in the state.
    """

    def __init__(self, process_noise: float = 0.0, obs_noise: float = 0.0, n_obs: int = 4) -> None:
        self.n_var = 4
        self.n_control_inputs = 1
        self.n_targets = 0
        self.n_dim = 1
        self.init_bounds = jnp.array([0.1, 0.1, 0.1, 0.1])
        super().__init__(process_noise, obs_noise, self.n_var, self.n_control_inputs, self.n_dim, n_obs)

        self.R = 0.1 * jnp.eye(self.n_control_inputs)

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Tuple[Array, Array]:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        x0 : :class:`jax.Array`
            Initial states.
        targets : :class:`jax.Array`
            Target states.
        """
        init_key, target_key = jrandom.split(key)
        x0 = jrandom.uniform(init_key, shape=(batch_size, self.n_var), minval=-self.init_bounds, maxval=self.init_bounds)
        targets = jnp.zeros((batch_size, self.n_targets))
        return x0, targets
    
    def sample_params(self, batch_size: int, mode: str, ts: Array, key: jrandom.PRNGKey) -> Tuple[Array, Array, Array, Array]:
        """
        Samples parameters for the environment.

        Parameters
        ----------
        batch_size : int
            Number of parameters to sample.
        mode : str
            Mode for sampling parameters.
        ts : :class:`jax.Array`
            Time steps.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        tuple of :class:`jax.Array`
            Sampled parameters.
        """
        l1 = l2 = m1 = m2 = jnp.ones((batch_size))
        return l1, l2, m1, m2
    
    def f_obs(self, key: jrandom.PRNGKey, t_x: Tuple[float, Array]) -> Tuple[jrandom.PRNGKey, Array]:
        """
        Computes the observation function.

        Parameters
        ----------
        key : :class:`jax.random.PRNGKey`
            Random key for sampling noise.
        t_x : tuple of (float, :class:`jax.Array`)
            Tuple containing the current time and state.

        Returns
        -------
        key : :class:`jax.random.PRNGKey`
            Updated random key.
        out : :class:`jax.Array`
            Observation.
        """
        _, out = super().f_obs(key, t_x)
        out = jnp.array([(out[0] + jnp.pi) % (2 * jnp.pi) - jnp.pi, (out[1] + jnp.pi) % (2 * jnp.pi) - jnp.pi, out[2], out[3]])[:self.n_obs]
        return key, out

    def initialize_parameters(self, params: Tuple[Array, Array, Array, Array], ts: Array) -> None:
        """
        Initializes the parameters of the environment.

        Parameters
        ----------
        params : tuple of :class:`jax.Array`
            Parameters to initialize.
        ts : :class:`jax.Array`
            Time steps.
        """
        l1, l2, m1, m2 = params
        self.l1 = l1  # [m]
        self.l2 = l2  # [m]
        self.m1 = m1  #: [kg] mass of link 1
        self.m2 = m2  #: [kg] mass of link 2
        self.lc1 = 0.5 * self.l1  #: [m] position of the center of mass of link 1
        self.lc2 = 0.5 * self.l2  #: [m] position of the center of mass of link 2
        self.moi1 = self.moi2 = 1.0
        self.g = 9.81

        self.G = jnp.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        self.V = self.process_noise * self.G

        self.C = jnp.eye(self.n_var)[:self.n_obs]
        self.W = self.obs_noise * jnp.eye(self.n_obs)
    
    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        control = jnp.squeeze(args)
        control = jnp.clip(control, -1, 1)
        theta1, theta2, theta1_dot, theta2_dot = state

        d1 = self.m1 * self.lc1**2 + self.m2 * (self.l1**2 + self.lc2**2 + 2 * self.l1 * self.lc2 * jnp.cos(theta2)) + self.moi1 + self.moi2
        d2 = self.m2 * (self.lc2**2 + self.l1 * self.lc2 * jnp.cos(theta2)) + self.moi2

        phi2 = self.m2 * self.lc2 * self.g * jnp.cos(theta1 + theta2 - jnp.pi/2)
        phi1 = -self.m2 * self.l1 * self.lc2 * theta2_dot**2 * jnp.sin(theta2) - 2 * self.m2 * self.l1 * self.lc2 * theta1_dot * theta2_dot * jnp.sin(theta1) \
                    + (self.m1 * self.lc1 + self.m2 * self.l1) * self.g * jnp.cos(theta1 - jnp.pi/2) + phi2
        
        if self.n_control_inputs == 1:
            theta2_acc = (control + d2/d1 * phi1 - self.m2 * self.l1 * self.lc2 * theta1_dot**2 * jnp.sin(theta2) - phi2) \
                        / (self.m2 * self.lc2**2 + self.moi2 - d2**2 / d1)
            theta1_acc = -(d2 * theta2_acc + phi1)/d1
        else:
            c1, c2 = control
            theta2_acc = (c1 + d2/d1 * phi1 - self.m2 * self.l1 * self.lc2 * theta1_dot**2 * jnp.sin(theta2) - phi2) \
                        / (self.m2 * self.lc2**2 + self.moi2 - d2**2 / d1)
            theta1_acc = (c2 - d2 * theta2_acc - phi1)/d1

        return jnp.array([
            theta1_dot,
            theta2_dot,
            theta1_acc,
            theta2_acc
        ])

    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.V

    def fitness_function(self, state: Array, control: Array, target: Array, ts: Array) -> float:
        """
        Computes the fitness function for the environment.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        control : :class:`jax.Array`
            Control inputs.
        target : :class:`jax.Array`
            Target states.
        ts : :class:`jax.Array`
            Time steps.

        Returns
        -------
        float
            Fitness value.
        """
        reached_threshold = jax.vmap(lambda theta1, theta2: -jnp.cos(theta1) - jnp.cos(theta1 + theta2) > 1.5)(state[:,0], state[:,1])
        first_success = jnp.argmax(reached_threshold)

        control = jnp.clip(control, -1, 1)

        control_cost = jax.vmap(lambda _state, _u: _u @ self.R @ _u)(state, control)
        costs = jnp.where((ts / (ts[1] - ts[0])) > first_success, jnp.zeros_like(control_cost), control_cost)

        return (first_success + (first_success == 0) * ts.shape[0] + jnp.sum(costs))
    
    def cond_fn_nan(self, t: float, y: Array, args: Tuple, **kwargs) -> float:
        """
        Checks for NaN or infinite values in the state.

        Parameters
        ----------
        t : float
            Current time.
        y : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        float
            -1.0 if NaN or infinite values are found, 1.0 otherwise.
        """
        return jnp.where((jnp.abs(y[2]) > (4 * jnp.pi)) | (jnp.abs(y[3]) > (9 * jnp.pi)) | jnp.any(jnp.isnan(y)) | jnp.any(jnp.isinf(y)), -1.0, 1.0)


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax
from typing import Tuple, Callable
from jaxtyping import Array
import jax.numpy as jnp
from kozax.fitness_functions.base_fitness_function import BaseFitnessFunction

class SymbolicRegressionFitnessFunction(BaseFitnessFunction):
    """
    Evaluator for candidates on symbolic regression tasks with x, y data.

    Methods
    -------
    __call__(candidate, data, tree_evaluator)
        Evaluates the candidate on a task.
    """

    def __call__(self, candidate: Array, data: Tuple[Array, Array], tree_evaluator: Callable) -> float:
        """
        Evaluates the candidate on a task.

        Parameters
        ----------
        candidate : :class:`jax.Array`
            The candidate solution to be evaluated.
        data : :class:`tuple` of :class:`jax.Array`
            The data required to evaluate the candidate. Tuple of (x, y) where x is the input data and y is the true output data.
        tree_evaluator : :class:`Callable`
            Function for evaluating trees.

        Returns
        -------
        float
            Fitness of the candidate.
        """
        x, y = data
        pred = jax.vmap(tree_evaluator, in_axes=[None, 0])(candidate, x)
        return jnp.mean(jnp.square(pred - y))


"""
kozax: Genetic programming framework in JAX

Copyright (c) 2024 sdevries0

This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivs 4.0 International License.
"""

import jax.numpy as jnp
import jax
import jax.random as jrandom
import diffrax
from kozax.environments.control_environments.control_environment_base import EnvironmentBase
from jaxtyping import Array
from typing import Tuple

class StirredTankReactor(EnvironmentBase):
    """
    Stirred Tank Reactor environment for control tasks.

    Parameters
    ----------
    process_noise : float
        Standard deviation of the process noise.
    obs_noise : float
        Standard deviation of the observation noise.
    n_obs : int, optional
        Number of observations. Default is 3.
    n_targets : int, optional
        Number of targets. Default is 1.
    max_control : :class:`jax.Array`, optional
        Maximum control values. Default is jnp.array([300]).
    external_f : callable, optional
        External influence function. Default is lambda t: 0.0.

    Attributes
    ----------
    n_var : int
        Number of variables in the state.
    n_control_inputs : int
        Number of control inputs.
    n_dim : int
        Number of dimensions.
    n_targets : int
        Number of targets.
    init_lower_bounds : :class:`jax.Array`
        Lower bounds for initial state sampling.
    init_upper_bounds : :class:`jax.Array`
        Upper bounds for initial state sampling.
    max_control : :class:`jax.Array`
        Maximum control values.
    Q : :class:`jax.Array`
        Process noise covariance matrix.
    r : :class:`jax.Array`
        Observation noise covariance matrix.
    external_f : callable
        External influence function.

    Methods
    -------
    initialize_parameters(params, ts)
        Initializes the parameters of the environment.
    sample_param_change(key, batch_size, ts, low, high)
        Samples parameter changes over time.
    sample_params(batch_size, mode, ts, key)
        Samples parameters for the environment.
    sample_init_states(batch_size, key)
        Samples initial states for the environment.
    drift(t, state, args)
        Computes the drift function for the environment.
    diffusion(t, state, args)
        Computes the diffusion function for the environment.
    fitness_function(state, control, targets, ts)
        Computes the fitness function for the environment.
    cond_fn_nan(t, y, args, **kwargs)
        Checks for NaN or infinite values in the state.
    """

    def __init__(self, process_noise: float = 0.0, obs_noise: float = 0.0, n_obs: int = 3, n_targets: int = 1, max_control: Array = jnp.array([300]), external_f: callable = lambda t: 0.0) -> None:
        self.process_noise = process_noise
        self.obs_noise = obs_noise
        self.n_var = 3
        self.n_control_inputs = 1
        self.n_dim = 1
        self.n_targets = n_targets
        self.init_lower_bounds = jnp.array([275, 350, 0.5])
        self.init_upper_bounds = jnp.array([300, 375, 1.0])
        self.max_control = max_control
        super().__init__(process_noise, obs_noise, self.n_var, self.n_control_inputs, self.n_dim, n_obs)

        self.Q = jnp.array([[0, 0, 0], [0, 0.01, 0], [0, 0, 0]])
        self.r = jnp.array([[0.0001]])
        self.external_f = external_f

    def initialize_parameters(self, params: Tuple[Array, Array, Array, Array, Array, Array, Array, Array], ts: Array) -> None:
        """
        Initializes the parameters of the environment.

        Parameters
        ----------
        params : tuple of :class:`jax.Array`
            Parameters to initialize.
        ts : :class:`jax.Array`
            Time steps.
        """
        Vol, Cp, dHr, UA, q, Tf, Tcf, Volc = params
        self.Ea = 72750     # activation energy J/gmol
        self.R = 8.314      # gas constant J/gmol/K
        self.k0 = 7.2e10    # Arrhenius rate constant 1/min
        self.Vol = Vol      # Volume [L]
        self.Cp = Cp        # Heat capacity [J/g/K]
        self.dHr = dHr      # Enthalpy of reaction [J/mol]
        self.UA = UA        # Heat transfer [J/min/K]
        self.q = q          # Flowrate [L/min]
        self.Cf = 1.0       # Inlet feed concentration [mol/L]
        self.Tf = diffrax.LinearInterpolation(ts, Tf)  # Inlet feed temperature [K]
        self.Tcf = Tcf      # Coolant feed temperature [K]
        self.Volc = Volc    # Cooling jacket volume

        self.k = lambda T: self.k0 * jnp.exp(-self.Ea / self.R / T)

        self.G = jnp.eye(self.n_var) * jnp.array([6, 6, 0.05])
        self.process_noise_ts = diffrax.LinearInterpolation(ts, jnp.linspace(self.process_noise[0], self.process_noise[1], ts.shape[0]))

        self.C = jnp.eye(self.n_var)[:self.n_obs]
        self.W = self.obs_noise * jnp.eye(self.n_obs) * (jnp.array([15, 15, 0.1])[:self.n_obs])

        self.max_control_ts = diffrax.LinearInterpolation(ts, jnp.hstack([mc * jnp.ones(int(ts.shape[0] // self.max_control.shape[0])) for mc in self.max_control]))
        self.external_influence = diffrax.LinearInterpolation(ts, jax.vmap(self.external_f)(ts))

    def sample_param_change(self, key: jrandom.PRNGKey, batch_size: int, ts: Array, low: float, high: float) -> Array:
        """
        Samples parameter changes over time.

        Parameters
        ----------
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.
        batch_size : int
            Number of samples.
        ts : :class:`jax.Array`
            Time steps.
        low : float
            Lower bound for sampling.
        high : float
            Upper bound for sampling.

        Returns
        -------
        :class:`jax.Array`
            Sampled parameter values.
        """
        init_key, decay_key = jrandom.split(key)
        decay_factors = jrandom.uniform(decay_key, shape=(batch_size,), minval=1.01, maxval=1.02)
        init_values = jrandom.uniform(init_key, shape=(batch_size,), minval=low, maxval=high)
        values = jax.vmap(lambda v, d, t: v * (d ** t), in_axes=[0, 0, None])(init_values, decay_factors, ts)
        return values

    def sample_params(self, batch_size: int, mode: str, ts: Array, key: jrandom.PRNGKey) -> Tuple[Array, Array, Array, Array, Array, Array, Array, Array]:
        """
        Samples parameters for the environment.

        Parameters
        ----------
        batch_size : int
            Number of samples.
        mode : str
            Sampling mode. Options are "Constant", "Different", "Changing".
        ts : :class:`jax.Array`
            Time steps.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        tuple of :class:`jax.Array`
            Sampled parameters.
        """
        if mode == "Constant":
            Vol = 100 * jnp.ones(batch_size)
            Cp = 239 * jnp.ones(batch_size)
            dHr = -5.0e4 * jnp.ones(batch_size)
            UA = 5.0e4 * jnp.ones(batch_size)
            q = 100 * jnp.ones(batch_size)
            Tf = 300 * jnp.ones((batch_size, ts.shape[0]))
            Tcf = 300 * jnp.ones(batch_size)
            Volc = 20.0 * jnp.ones(batch_size)
        elif mode == "Different":
            keys = jrandom.split(key, 8)
            Vol = jrandom.uniform(keys[0], (batch_size,), minval=75, maxval=150)
            Cp = jrandom.uniform(keys[1], (batch_size,), minval=200, maxval=350)
            dHr = jrandom.uniform(keys[2], (batch_size,), minval=-55000, maxval=-45000)
            UA = jrandom.uniform(keys[3], (batch_size,), minval=25000, maxval=75000)
            q = jrandom.uniform(keys[4], (batch_size,), minval=75, maxval=125)
            Tf = jnp.repeat(jrandom.uniform(keys[5], (batch_size,), minval=300, maxval=350)[:, None], ts.shape[0], axis=1)
            Tcf = jrandom.uniform(keys[6], (batch_size,), minval=250, maxval=300)
            Volc = jrandom.uniform(keys[7], (batch_size,), minval=10, maxval=30)
        elif mode == "Changing":
            keys = jrandom.split(key, 8)
            Vol = jrandom.uniform(keys[0], (batch_size,), minval=75, maxval=150)
            Cp = jrandom.uniform(keys[1], (batch_size,), minval=200, maxval=350)
            dHr = jrandom.uniform(keys[2], (batch_size,), minval=-55000, maxval=-45000)
            UA = jrandom.uniform(keys[3], (batch_size,), minval=25000, maxval=75000)
            q = jrandom.uniform(keys[4], (batch_size,), minval=75, maxval=125)
            Tf = self.sample_param_change(keys[5], batch_size, ts, 300, 350)
            Tcf = jrandom.uniform(keys[6], (batch_size,), minval=250, maxval=300)
            Volc = jrandom.uniform(keys[7], (batch_size,), minval=10, maxval=30)
        return Vol, Cp, dHr, UA, q, Tf, Tcf, Volc

    def sample_init_states(self, batch_size: int, key: jrandom.PRNGKey) -> Tuple[Array, Array]:
        """
        Samples initial states for the environment.

        Parameters
        ----------
        batch_size : int
            Number of initial states to sample.
        key : :class:`jax.random.PRNGKey`
            Random key for sampling.

        Returns
        -------
        x0 : :class:`jax.Array`
            Initial states.
        targets : :class:`jax.Array`
            Target states.
        """
        init_key, target_key = jrandom.split(key)
        x0 = jrandom.uniform(init_key, shape=(batch_size, self.n_var), minval=self.init_lower_bounds, maxval=self.init_upper_bounds)
        targets = jrandom.uniform(target_key, shape=(batch_size, self.n_targets), minval=400, maxval=480)
        return x0, targets
    
    def drift(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the drift function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Drift.
        """
        Tc, T, c = state
        control = jnp.squeeze(args)
        control = jnp.clip(control, 0, 300)

        dc = (self.q / self.Vol) * (self.Cf - c) - self.k(T) * c
        dT = (self.q / self.Vol) * (self.Tf.evaluate(t) - T) + (-self.dHr / self.Cp) * self.k(T) * c + (self.UA / self.Vol / self.Cp) * (Tc - T) + self.external_influence.evaluate(t)
        dTc = (control / self.Volc) * (self.Tcf - Tc) + (self.UA / self.Volc / self.Cp) * (T - Tc)

        return jnp.array([dTc, dT, dc])

    def diffusion(self, t: float, state: Array, args: Tuple) -> Array:
        """
        Computes the diffusion function for the environment.

        Parameters
        ----------
        t : float
            Current time.
        state : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.

        Returns
        -------
        :class:`jax.Array`
            Diffusion.
        """
        return self.process_noise_ts.evaluate(t) * self.G

    def fitness_function(self, state: Array, control: Array, targets: Array, ts: Array) -> float:
        """
        Computes the fitness function for the environment.

        Parameters
        ----------
        state : :class:`jax.Array`
            Current state.
        control : :class:`jax.Array`
            Control inputs.
        targets : :class:`jax.Array`
            Target states.
        ts : :class:`jax.Array`
            Time steps.

        Returns
        -------
        float
            Fitness value.
        """
        x_d = jax.vmap(lambda tar: jnp.array([0, tar, 0]))(targets)
        costs = jax.vmap(lambda _state, _u, _x_d: (_state - _x_d).T @ self.Q @ (_state - _x_d) + (_u) @ self.r @ (_u))(state, control, x_d)
        return jnp.sum(costs)

    def cond_fn_nan(self, t: float, y: Array, args: Tuple, **kwargs) -> float:
        """
        Checks for NaN or infinite values in the state.

        Parameters
        ----------
        t : float
            Current time.
        y : :class:`jax.Array`
            Current state.
        args : tuple
            Additional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        float
            -1.0 if NaN or infinite values are found, 1.0 otherwise.
        """
        return jnp.where(jnp.any(jnp.isinf(y) + jnp.isnan(y)), -1.0, 1.0)